# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
import numpy as np

from msprobe.msaccucmp.cmp_utils import log
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError
from msprobe.msaccucmp.dump_parse.dump_data_object import DumpTensor, DumpDataObj
from msprobe.msaccucmp.dump_parse import dump


class FFTSParser:
    """
    The class for FFTS mode type parser
    """
    def __init__(self, dump_file_list, dump_data_list):
        self.dump_file_list = dump_file_list
        self.dump_data_list = dump_data_list
    
    @property
    def parse_ffts(self: any) -> tuple:
        """
        parse the ffts mode dump data and merge data
        @return: file path, dump data
        """
        if len(self.dump_file_list) == 1:
            dump_data = self.dump_data_list[-1]
            file_path = self.dump_file_list[-1]
            dump_data.set_op_attr(dump_data.op_name, True)
            return file_path, dump_data
        dump_base = self.dump_data_list[0]
        thread_num = dump_base.get_thread_num
        if self.check_file_missing(thread_num):
            dump_base.ffts_file_check = False
            log.print_warn_log(
                f"This is a FFTS+ mode dump data {dump_base.op_name},"
                f" The number of files does not match the number of thread (instance slice num).")

        cut_axis = dump_base.get_cut_axis_auto if dump_base.get_ffts_mode else dump_base.get_cut_axis_manual

        if not cut_axis or self.check_invalid_cut_axis(cut_axis):
            dump_data_to_file = list(zip(self.dump_data_list, self.dump_file_list))
            dump_data_to_file.sort(key=lambda x: os.path.basename(x[1]).split(".")[4])
            file_path = dump_data_to_file[-1][1]
            dump_data = dump_data_to_file[-1][0]
            log.print_warn_log("The cut axis of Dump data is invalid. The current compare dump file is {}. "
                               "All dump files are {}".format(dump_data_to_file[-1][1], ",".join(self.dump_file_list)))
        else:
            output_num = len(dump_base.output_data)
            if dump_base.get_ffts_mode:
                output_data_list = []
                for index, dump_file in enumerate(self.dump_data_list):
                    output_shape = dump_file.calculate_auto_mode_shape(index, "output")
                    output_data_list.append(dump_file.get_auto_output_data(output_shape))
            else:
                output_data_list = [dump_data.get_output_data() for dump_data in self.dump_data_list]

            expected_len = len(output_data_list[0])
            for output in (output_data_list):
                if len(output) != expected_len:
                    log.print_error_log(
                        f"Inconsistent output length detected: expected {expected_len}, "
                        f"but got {len(output)}"
                    )
                    raise CompareError(CompareError.MSACCUCMP_INVALID_DUMP_DATA_ERROR)

            dump_data_output_list = []
            for i in range(output_num):
                output_index = []
                for output in output_data_list:
                    output_index.append(output[i])
                dump_data_output_list.append(output_index)

            merge_output = self.merge_data(dump_data_output_list, cut_axis)
            dump_data = self.create_merge_dump_data(dump_base, merge_output)
            file_path = '.'.join(self.dump_file_list[0].split(".")[:4] + ['*'])
            log.print_info_log(f"This is a FFTS+ mode dump data {dump_base.op_name}, "
                               f"output data has been merged, new file path is {file_path}")
        return file_path, dump_data

    @staticmethod
    def check_invalid_cut_axis(cut_axis: list) -> bool:
        """
        check if the cut axis is valid
        @param cut_axis: cut axis
        @return: True or False
        """
        return all(dim == [] for dim in cut_axis)

    @staticmethod
    def merge_data(output_list: list, cut_axis: list) -> list:
        merge_output = []
        for index, dim in enumerate(cut_axis):
            if not dim:
                merge_output.append(output_list[index][0])
            else:
                axis = cut_axis[index][0]
                merge_output.append(np.concatenate(output_list[index], axis))
        return merge_output

    @staticmethod
    def create_merge_dump_data(dump_base: DumpDataObj, merge_output: list) -> DumpDataObj:
        dump_data = DumpDataObj()
        op_name = dump.process_op_name(dump_base.op_name) if dump_base.op_name else ""
        dump_data.set_op_attr(op_name, dump_base.ffts_file_check)
        for index, data in enumerate(merge_output):
            shape = list(data.shape)
            common_attr = dump_base.output_data[index].get_common_attr
            dump_tensor = DumpTensor(index=index, data=data.reshape(-1), shape=shape,
                                     data_type=common_attr[0], tensor_format=common_attr[1],
                                     address=common_attr[2], original_shape=common_attr[3], is_ffts=True)
            dump_data.output_data.append(dump_tensor)
        dump_data.input_data = dump_base.input_data
        return dump_data

    def check_file_missing(self, thread_num: int) -> bool:
        """
        check if file number match thread number
        @param thread_num: thread number
        @return: True of False
        """
        return len(self.dump_data_list) != thread_num



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from typing import List, Dict, Union, Any

import numpy as np

from msprobe.core.overflow_check.api_info import APIInfo
from msprobe.core.overflow_check.level import OverflowLevel
from msprobe.core.overflow_check.utils import has_nan_inf
from msprobe.core.common.decorator import recursion_depth_decorator


class AnomalyScene:
    """异常场景的基类"""

    def __init__(self, api_info: APIInfo):
        self.api_name = api_info.api_name
        self.api_data = api_info

    @property
    def rank(self) -> OverflowLevel:
        """获取异常等级"""
        raise NotImplementedError

    @staticmethod
    @recursion_depth_decorator("AbnormalScene: AnomalyScene._has_anomaly")
    def _has_anomaly(data: Union[Dict, Any]) -> bool:
        """检查张量是否包含异常值"""
        if isinstance(data, dict):
            return has_nan_inf(data)
        elif isinstance(data, list):
            return any(AnomalyScene._has_anomaly(x) for x in data)
        return False

    def get_details(self) -> Dict:
        """获取异常详情"""
        return {
            'api_name': self.api_name,
            'rank': self.rank.value,
            'scene_type': self.__class__.__name__,
            'input_args_anomaly_indices': self._get_anomaly_indices_from_list(self.api_data.input_args),
            'input_kwargs_anomaly_keys': self._get_anomaly_keys_from_dict(self.api_data.input_kwargs),
            'output_anomaly_indices': self._get_anomaly_indices_from_list(self.api_data.output_data)
        }

    def matches(self) -> bool:
        """
            待子类实现对应匹配逻辑
        Returns:

        """
        raise NotImplementedError

    def _get_anomaly_indices_from_list(self, data_list: List[Dict]) -> List[int]:
        return [i for i, data in enumerate(data_list) if self._has_anomaly(data)]

    def _get_anomaly_keys_from_dict(self, data_dict: Dict) -> List[str]:
        return [key for key, data in data_dict.items() if self._has_anomaly(data)]


class InputOutputAnomalyScene(AnomalyScene):
    """输入输出异常检测的基类"""
    def has_input_anomaly(self) -> bool:
        """检查输入是否有异常（包括args和kwargs）"""
        # args
        args_anomaly = any(self._has_anomaly(x) for x in self.api_data.input_args)
        # kwargs
        kwargs_anomaly = any(self._has_anomaly(x) for x in self.api_data.input_kwargs.values())
        return args_anomaly or kwargs_anomaly

    def has_output_anomaly(self) -> bool:
        """检查输出是否有异常"""
        return any(self._has_anomaly(x) for x in self.api_data.output_data)

    def matches(self) -> bool:
        """判断是否匹配该场景"""
        raise NotImplementedError


class InputAnomalyOutputNormalScene(InputOutputAnomalyScene):
    """输入异常，输出正常场景"""

    @property
    def rank(self) -> OverflowLevel:
        return OverflowLevel.MEDIUM

    def matches(self) -> bool:
        return self.has_input_anomaly() and not self.has_output_anomaly()


class InputAnomalyOutputAnomalyScene(InputOutputAnomalyScene):
    """输入异常，输出异常场景"""

    @property
    def rank(self) -> OverflowLevel:
        return OverflowLevel.HIGH

    def matches(self) -> bool:
        return self.has_input_anomaly() and self.has_output_anomaly()


class InputNormalOutputAnomalyScene(InputOutputAnomalyScene):
    """输入正常，输出异常场景"""

    @property
    def rank(self) -> OverflowLevel:
        return OverflowLevel.CRITICAL

    def matches(self) -> bool:
        return not self.has_input_anomaly() and self.has_output_anomaly()


class NumericalMutationScene(AnomalyScene):
    """
        检查数值突变，统计输入args、kwargs中norm值，同时统计输出的norm最大值，计算差异，大于 threshold 则认为是异常情况
    """
    def __init__(self, api_info: APIInfo, threshold: float = 100.0):
        super().__init__(api_info)
        self.threshold = threshold

    @property
    def rank(self) -> OverflowLevel:
        return OverflowLevel.HIGH

    @staticmethod
    def _get_tensor_norms(data_list: List[Dict]) -> List[float]:
        norms = []
        for data in data_list:
            if isinstance(data, dict) and data.get('type') == 'torch.Tensor':
                norm = data.get('Norm')
                if norm is not None and not np.isnan(norm):
                    norms.append(norm)
        return norms

    @staticmethod
    def _get_kwargs_norms(data_dict: Dict) -> List[float]:
        """
            获取kwargs中张量的范数列表
        Args:
            data_dict:
        Returns:
        """
        norms = []
        for data in data_dict.values():
            if isinstance(data, dict) and data.get('type') == 'torch.Tensor':
                norm = data.get('Norm')
                if norm is not None and not np.isnan(norm):
                    norms.append(norm)
        return norms

    def matches(self) -> bool:
        """
            继承父类函数，实现数值突变检查
        Returns:
        """
        # 收集所有输入的范数
        input_norms = (self._get_tensor_norms(self.api_data.input_args) +
                       self._get_kwargs_norms(self.api_data.input_kwargs))
        # 收集所有输出的范数
        output_norms = self._get_tensor_norms(self.api_data.output_data)

        if not input_norms or not output_norms:
            return False

        max_input = max(input_norms)
        max_output = max(output_norms)

        if max_input == 0:
            return max_output > self.threshold
        return max_output / max_input > self.threshold

    def get_details(self) -> Dict:
        details = super().get_details()
        details.update({
            'threshold': self.threshold,
            'scale_change_detected': self.matches()
        })
        return details



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from __future__ import annotations

from typing import Any, Dict, Iterable, Optional, Tuple

from msprobe.core.common.const import MonitorConst
from msprobe.core.common.framework_adapter import FmkAdp
from msprobe.core.common.log import logger


def iter_model_chunks(model: Any) -> Iterable[Tuple[Any, str]]:
    """
    Iterate over model or model_list, yielding (model_chunk, prefix).

    - Single model: yields (model, "")
    - List/tuple: yields (sub_model, "<idx><NAME_SEP>")
    """
    if model is None:
        # 保持与后续分支一致的生成器行为
        return
        # 不再显式 return ()

    if isinstance(model, (list, tuple)):
        for idx, sub in enumerate(model):
            if sub is None:
                continue
            prefix = f"{idx}{MonitorConst.NAME_SEP}"
            yield sub, prefix
        return
        # 同样去掉显式的 return ()

    yield model, ""


def iter_model_params(model: Any) -> Iterable[Tuple[str, Any]]:
    """
    Iterate (name, param) pairs for a model that may be:
      - a single nn.Module/nn.Cell
      - a list/tuple of models

    For list/tuple, each sub-model is prefixed with
    "<idx><NAME_SEP>" to keep names unique and consistent
    with other multi-model monitor behaviours.
    """
    for chunk, prefix in iter_model_chunks(model):
        try:
            for name, param in FmkAdp.named_parameters(chunk):
                yield prefix + name, param
        except Exception:
            # If a sub-element is not a valid module, skip it.
            continue


def iter_model_modules(model: Any) -> Iterable[Tuple[str, Any]]:
    """
    Iterate (name, module) pairs for a model or model_list, using
    framework-specific module traversal, with optional vpp_stage prefix.
    """
    for chunk, prefix in iter_model_chunks(model):
        try:
            for name, module in FmkAdp.iter_named_modules(chunk):
                yield prefix + name, module
        except Exception:
            logger.warning("Failed to iterate modules for model chunk, skipping.")
            continue


def build_param2name(model: Any = None, optimizer: Any = None) -> Dict[Any, str]:
    """
    Build a param->name mapping from model/optimizer.

    - Primary source: model (or list/tuple of models), using framework
      specific named_parameters().
    - Fallback: optimizer.param_groups when model mapping is empty.
    """
    param2name: Dict[Any, str] = {}

    for name, param in iter_model_params(model):
        param2name[param] = name

    if param2name or optimizer is None:
        return param2name

    idx = 0
    for group in getattr(optimizer, "param_groups", []):
        for param in group.get("params", []):
            if param not in param2name:
                param2name[param] = f"param_{idx}"
                idx += 1

    return param2name


def get_vpp_stage_from_tag(tag: str) -> Optional[int]:
    """
    Extract vpp_stage index from a tag or name that may start with
    "<idx><NAME_SEP>". Returns None if no valid stage prefix exists.
    """
    if not tag:
        return None
    prefix, sep, _ = tag.partition(MonitorConst.NAME_SEP)
    if not sep:
        return None
    try:
        return int(prefix)
    except (TypeError, ValueError):
        return None



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
import time
import sys
from functools import wraps
from msprobe.core.common.const import MsgConst


def filter_special_chars(func):
    @wraps(func)
    def func_level(self, msg, **kwargs):
        for char in MsgConst.SPECIAL_CHAR:
            msg = msg.replace(char, '_')
        return func(self, msg, **kwargs)

    return func_level


class BaseLogger:
    def __init__(self):
        self.rank = None
        self.level = self.get_level()

    @staticmethod
    def get_level():
        input_level = os.environ.get(MsgConst.MSPROBE_LOG_LEVEL)
        if input_level not in MsgConst.LOG_LEVEL_ENUM:
            return MsgConst.LogLevel.INFO.value
        else:
            return int(input_level)

    @filter_special_chars
    def raw(self, msg, **kwargs):
        """
        直接输出原始内容，不添加日志前缀
        用于输出子进程的原始日志，避免双份前缀
        """
        msg = msg.rstrip("_")
        print(msg)
        sys.stdout.flush()

    def get_rank(self):
        return self.rank

    @filter_special_chars
    def error(self, msg):
        if self.level <= MsgConst.LogLevel.ERROR.value:
            self._print_log(MsgConst.LOG_LEVEL[3], msg)

    @filter_special_chars
    def warning(self, msg):
        if self.level <= MsgConst.LogLevel.WARNING.value:
            self._print_log(MsgConst.LOG_LEVEL[2], msg)

    @filter_special_chars
    def info(self, msg):
        if self.level <= MsgConst.LogLevel.INFO.value:
            self._print_log(MsgConst.LOG_LEVEL[1], msg)

    @filter_special_chars
    def debug(self, msg):
        if self.level <= MsgConst.LogLevel.DEBUG.value:
            self._print_log(MsgConst.LOG_LEVEL[0], msg)

    def on_rank_0(self, func):
        def func_rank_0(*args, **kwargs):
            current_rank = self.get_rank()
            if current_rank is None or current_rank == 0:
                return func(*args, **kwargs)
            else:
                return None

        return func_rank_0

    def info_on_rank_0(self, msg):
        return self.on_rank_0(self.info)(msg)

    def error_on_rank_0(self, msg):
        return self.on_rank_0(self.error)(msg)

    def warning_on_rank_0(self, msg):
        return self.on_rank_0(self.warning)(msg)

    def error_log_with_exp(self, msg, exception):
        self.error(msg)
        raise exception

    def warning_log_with_exp(self, msg, exception):
        """
        打印警告日志并抛出指定异常
        """
        self.warning(msg)
        raise exception

    def _print_log(self, level, msg, end='\n'):
        current_rank = self.get_rank()
        current_time = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        pid = os.getpid()
        if current_rank is not None:
            full_msg = f"{current_time} ({pid}) [rank {current_rank}] [{level}] {msg}"
        else:
            full_msg = f"{current_time} ({pid}) [{level}] {msg}"
        print(full_msg, end=end)
        sys.stdout.flush()


logger = BaseLogger()



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This file mainly involves the print function.
"""

import os
import logging
from functools import wraps
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager


def filter_special_chars(func):
    @wraps(func)
    def func_level(msg, **kwargs):
        if isinstance(msg, str):
            for char in ConstManager.SPECIAL_CHAR:
                msg = msg.replace(char, '_')
        return func(msg, **kwargs)

    return func_level


def _setting_config(message: str) -> str:
    pid = os.getpid()
    cur_format = '%(asctime)s (' + str(pid) + ') - [%(levelname)s] %(message)s'
    logging.basicConfig(format=cur_format, datefmt="%Y-%m-%d %H:%M:%S",
                        level=logging.INFO)
    return message


@filter_special_chars
def print_error_log(error_msg: str) -> None:
    """
    print error log
    :param error_msg: the error message
    """
    message = _setting_config(error_msg)
    logging.error(message)


@filter_special_chars
def print_warn_log(warn_msg: str) -> None:
    """
    print warn log
    :param warn_msg: the warn message
    """
    message = _setting_config(warn_msg)
    logging.warning(message)


@filter_special_chars
def print_info_log(info_msg: str) -> None:
    """
    print info log
    :param info_msg: the info message
    """
    message = _setting_config(info_msg)
    logging.info(message)


def print_no_left_dump_file_error(op_name: str, op_type: str, is_error: bool = False) -> str:
    """
    Print warn or error log for no my output dump file error
    :param op_name: the op name
    :param op_type: the op type
    :param is_error: the log lever is error
    :return: message
    """
    message = '[%s] There is no dump file for my output operator "%s". The type is %s.' \
          % (op_name, op_name, op_type)
    if is_error:
        print_error_log(message)
    else:
        print_warn_log(message)
    return message


def print_no_right_dump_file_error(op_name: str, tensor_id: str, is_error: bool = False) -> str:
    """
    Print warn or error log for no right dump file error
    :param op_name: the op name
    :param tensor_id: the tensor id
    :param is_error: the log lever is error
    :return message
    """
    message = '[%s] There is no the ground truth dump file for %s.' % (op_name, tensor_id)
    if is_error:
        print_error_log(message)
    else:
        print_warn_log(message)
    return message


def print_invalid_nz_dump_data(message: str, op_name: str = None, is_error: bool = False) -> str:
    if not is_error and op_name:
        message = "[%s] %s" % (op_name, message)
    if is_error:
        print_error_log(message)
    else:
        print_warn_log(message)
    return message


def print_start_to_compare_op(op_name: str) -> None:
    """
    Print info log for start to compare op
    :param op_name: the op name
    """
    message = '[%s] Start to compare op "%s".' % (op_name, op_name)
    print_info_log(message)


def print_open_file_error(path: str, io_error: any) -> None:
    """
    Print error log for open file error
    :param path: the path
    :param io_error: error info
    """
    message = 'Failed to open "%r". %s' % (path, str(io_error))
    print_error_log(message)


def print_write_result_info(prefix: str, path: str) -> None:
    """
    Print info log for write result to file
    :param prefix: the info
    :param path: the path
    """
    message = 'The %s have been written to "%r".' % (prefix, path)
    print_info_log(message)


def print_only_support_error(prefix: str, value: any, support_info: list) -> None:
    """
    Print error log for only supports error
    :param prefix: the info
    :param value: the value no support
    :param support_info: the support info
    """
    message = "The %s '%s' is invalid. It only supports '%s'." % (prefix, str(value), str(support_info))
    print_error_log(message)


def print_not_match_error(op_name: str, prefix: str, left_value: str, right_value: str, tensor_id: str = None) -> str:
    """
    Print not match error
    :param op_name: the op name
    :param prefix: the info
    :param left_value: the left value
    :param right_value: the right value
    :param tensor_id: the tensor id
    :return message
    """
    line = '[%s] The %s does not match (%s vs %s)' % (op_name, prefix, left_value, right_value)
    if tensor_id:
        line = '%s for %s.' % (line, tensor_id)
        print_error_log(line)
    else:
        line = '%s.' % line
        print_warn_log(line)
    return line


def print_cannot_compare_warning(op_name: str, left_shape: str, right_shape: str) -> str:
    """
    Print cannot compare warning
    :param op_name: the op name
    :param left_shape: the left_shape
    :param right_shape: the right shape
    :return message
    """
    prefix = '[%s] ' % op_name if op_name else ''
    message = '%sDue to the different shapes on the left and right,the left dump data%s can not ' \
              'be compared to the right dump data%s. Please check the batch of the dump data or ' \
              'the shape may be changed due to optimization.' % (prefix, left_shape, right_shape)
    print_warn_log(message)
    return message


def print_npu_path_valid_message(npu_dump_dir: str, dump_file_path_format: str) -> str:
    """
    Print npu path valid message
    :param npu_dump_dir: the npu dump directory
    :param dump_file_path_format : correct dump file path format
    :return message
    """
    message = "The {0} does not match the path format," \
              "please save dump files in the {1} path format".format(npu_dump_dir, dump_file_path_format)
    print_error_log(message)
    return message


def print_out_of_range_error(op_name: str, index_type: str, index: int, range_str: str) -> None:
    """
    Print out of range error
    :param op_name: the op name
    :param index_type: the tensor type
    :param index: the index
    :param range_str: the count
    """
    prefix = ''
    if op_name:
        prefix = '[%s] ' % op_name
    message = '%sThe %s index (%d) is out of range %s. Please check the index.' % \
          (prefix, index_type, index, range_str)
    print_error_log(message)


def print_skip_inner_op_msg(op_name: str, is_error: bool) -> None:
    """
    Print warn or error log for skip inner operator
    :param op_name: the op name
    :param is_error: the log lever is error
    :return message
    """
    message = '[%s] The op "%s" is inner node for multi to multi relation. Skip the op "%s".' \
          % (op_name, op_name, op_name)
    if is_error:
        print_error_log(message)
    else:
        print_warn_log(message)


def print_deprecated_warning(file_name: str) -> None:
    """
    Print deprecated warning
    :param file_name: the file name
    """
    message = 'Note that "%s" will be deprecated in a future release. It'\
              ' is recommended to use the next-generation "msaccucmp.py".' % file_name
    print_warn_log(message)


def print_skip_quant_info(op_name: str) -> None:
    """
    Print the op skipped info
    :param op_name: the op name
    """
    message = '[%s] This op is in a quant/dequant op pair. Skip the op.' % op_name
    print_info_log(message)



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import setuptools

VERSION = '8.3.2'
INSTALL_REQUIRED = ["tensorboard >= 2.11.2"]

setuptools.setup(
    name="tb-graph-ascend",
    version=VERSION,
    description="Model Hierarchical Visualization TensorBoard Plugin",
    long_description="Model Hierarchical Visualization TensorBoard Plugin : \
        https://gitcode.com/Ascend/mstt/tree/master/plugins/tensorboard-plugins/tb_graph_ascend",
    url="https://gitcode.com/Ascend/mstt/tree/master/plugins/tensorboard-plugins/tb_graph_ascend",
    author="Ascend Team",
    author_email="pmail_mindstudio@huawei.com",
    packages=setuptools.find_packages(),
    package_data={
        "server": ["static/**"],
    },
    entry_points={
        "tensorboard_plugins": [
            "graph_ascend = server.plugin:GraphsPlugin",
        ],
    },
    python_requires=">=3.7",
    install_requires=INSTALL_REQUIRED,
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    license='Mulan PSL v2',
    keywords='tensorboard graph ascend plugin',
)


# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os

from msprobe.core.common.const import Const
from msprobe.core.common.file_utils import load_json
from msprobe.core.dump.common_config import BaseConfig, CommonConfig
from msprobe.pytorch.dump.api_dump.utils import get_ops
from msprobe.core.common.utils import is_int


class TensorConfig(BaseConfig):
    def __init__(self, json_config):
        super().__init__(json_config)
        self.check_config()
        self._check_summary_mode()
        self._check_file_format()
        self.bench_path = json_config.get("bench_path")

    def _check_file_format(self):
        if self.file_format is not None and self.file_format not in ["npy", "bin"]:
            raise Exception("file_format is invalid")


class StatisticsConfig(BaseConfig):
    def __init__(self, json_config):
        super().__init__(json_config)
        self.check_config()
        self._check_summary_mode()

        self.tensor_list = json_config.get("tensor_list", [])
        self._check_str_list_config(self.tensor_list, "tensor_list")


class ACCCheckConfig(BaseConfig):
    WrapApi = get_ops()

    def __init__(self, json_config):
        super().__init__(json_config)
        self.white_list = json_config.get("white_list", Const.DEFAULT_LIST)
        self.black_list = json_config.get("black_list", Const.DEFAULT_LIST)
        self.error_data_path = json_config.get("error_data_path", Const.DEFAULT_PATH)

        self.check_acc_check_config()

    @classmethod
    def check_filter_list_config(cls, key, filter_list):
        if not isinstance(filter_list, list):
            raise Exception("%s must be a list type" % key)
        if not all(isinstance(item, str) for item in filter_list):
            raise Exception("All elements in %s must be string type" % key)
        invalid_api = [item for item in filter_list if item not in cls.WrapApi]
        if invalid_api:
            raise Exception("Invalid api in %s: %s" % (key, invalid_api))

    @classmethod
    def check_error_data_path_config(cls, error_data_path):
        if not os.path.exists(error_data_path):
            raise Exception("error_data_path: %s does not exist" % error_data_path)

    def check_acc_check_config(self):
        ACCCheckConfig.check_filter_list_config(Const.WHITE_LIST, self.white_list)
        ACCCheckConfig.check_filter_list_config(Const.BLACK_LIST, self.black_list)
        ACCCheckConfig.check_error_data_path_config(self.error_data_path)


class DiffCheckConfig(BaseConfig):
    def __init__(self, json_config):
        super().__init__(json_config)
        self.diff_nums = json_config.get("diff_nums")
        self.check_mode = json_config.get("check_mode")
        self.check_diff_config()

    def check_diff_config(self):
        if self.diff_nums is not None and not is_int(self.diff_nums):
            raise Exception("diff_num is invalid")
        if self.diff_nums is not None and self.diff_nums != -1 and self.diff_nums <= 0:
            raise Exception("diff_nums should be -1 or positive integer")
        if self.check_mode is not None and self.check_mode not in ["all", "aicore", "atomic"]:
            raise Exception("check_mode is invalid")


class StructureConfig(BaseConfig):
    def __init__(self, json_config):
        super().__init__(json_config)


TaskDict = {
    Const.TENSOR: TensorConfig,
    Const.STATISTICS: StatisticsConfig,
    Const.DIFF_CHECK: DiffCheckConfig,
    Const.STRUCTURE: StructureConfig,
    Const.ACC_CHECK: ACCCheckConfig
}


def parse_task_config(task, json_config):
    if task not in Const.TORCH_TASK_LIST:
        raise Exception(f"task <{task}> is invalid, it is not in the {Const.TORCH_TASK_LIST}.")
    task_map = json_config.get(task, dict())
    return TaskDict.get(task)(task_map)


def parse_json_config(json_file_path, task):
    if not json_file_path:
        config_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        json_file_path = os.path.join(config_dir, "../../config.json")
    json_config = load_json(json_file_path)
    common_config = CommonConfig(json_config)
    if task:
        task_config = parse_task_config(task, json_config)
    else:
        task_config = parse_task_config(common_config.task, json_config)
    return common_config, task_config



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This file mainly involves the common function.
"""
import itertools
import multiprocessing
import os
import uuid

import numpy as np

from msprobe.msaccucmp.cmp_utils import log, path_check
from msprobe.msaccucmp.dump_parse import dump_utils, mapping
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager
from msprobe.msaccucmp.vector_cmp.compare_detail.detail import DetailInfo
from msprobe.msaccucmp.cmp_utils.file_utils import FileUtils
from msprobe.msaccucmp.cmp_utils.multi_process.progress import Progress


class MinMaxValue:
    """
    The class for absolute and relative value min max value recorder
    """

    def __init__(self: any) -> None:
        self._min_absolute_error = np.nan
        self._max_absolute_error = np.nan
        self._min_relative_error = np.nan
        self._max_relative_error = np.nan

    def get_min_and_max_value(self: any) -> str:
        """
        Get min/max absolute error and min/max relative error
        """
        return "MinAbsoluteError:%.6f\nMaxAbsoluteError:%.6f\nMinRelativeError:%.6f\nMaxRelativeError:%.6f" \
               % (self._min_absolute_error, self._max_absolute_error,
                  self._min_relative_error, self._max_relative_error)

    def set_min_absolute_error(self: any, value: any) -> None:
        """
        Set min absolute error
        """
        self._min_absolute_error = value

    def set_max_absolute_error(self: any, value: any) -> None:
        """
        Set max absolute error
        """
        self._max_absolute_error = value

    def set_min_relative_error(self: any, value: any) -> None:
        """
        Set min relative error
        """
        self._min_relative_error = value

    def set_max_relative_error(self: any, value: any) -> None:
        """
        Set max relative error
        """
        self._max_relative_error = value


class TopN:
    """
    The class for recording the top n absolute/relative error values
    """

    def __init__(self: any, is_bool: bool = False) -> None:
        self._absolute_error_top_n_list = []
        self._relative_error_top_n_list = []
        self.is_bool = is_bool

    def set_absolute_top_n_list(self: any, index_list: list, *data_lists: any) -> None:
        """
        set and store the top n absolute error values
        """
        self._absolute_error_top_n_list = self._make_top_n_list(index_list, *data_lists)

    def set_relative_top_n_list(self: any, index_list: list, *data_lists: any) -> None:
        """
        set and store the top n relative error values
        """
        self._relative_error_top_n_list = self._make_top_n_list(index_list, *data_lists)

    def get_absolute_error_top_n_list(self: any) -> list:
        """
        get top n absolute error list
        """
        return self._absolute_error_top_n_list

    def get_relative_error_top_n_list(self: any) -> list:
        """
        get top n relative error list
        """
        return self._relative_error_top_n_list

    def _make_top_n_list(self: any, index_list: list, *data_lists: any) -> list:
        if len(data_lists) < 6:
            raise RuntimeError('The number of arguments is not correct.')
        dim_list = data_lists[0]
        left_list = data_lists[1]
        right_list = data_lists[2]
        absolute_error_list = data_lists[3]
        relative_error_list = data_lists[4]
        error_type = data_lists[5]
        top_n_original_list = []
        for i, index in enumerate(index_list):
            # variable row value,for example,[17866338,"22 16 32 98",-0.866699,-0.000017,0.866682,51349.906250]
            row = [
                index, " ".join((str(x) for x in dim_list[i])), left_list[i], right_list[i],
                absolute_error_list[i], relative_error_list[i]
            ]
            top_n_original_list.append(row)
        if error_type == "absolute":
            # 4 is AbsoluteError column subscript.
            # 5 is RelativeError column subscript.
            # 0 is Index column subscript.
            return self._get_top_n_result(top_n_original_list, 4, 0)
        return self._get_top_n_result(top_n_original_list, 5, 0)

    def _get_top_n_result(self: any, original_list: list, first_key: int, second_key: int) -> list:
        top_n_list = []
        sort_list = sorted(original_list, key=lambda x: (x[first_key], x[second_key]), reverse=True)
        if not self.is_bool:
            item_template = '%d,%s,%.6f\t,%.6f\t,%.6f\t,%.6f\t\n'
        else:
            item_template = '%d,%s,%s\t,%s\t,%.6f\t,%.6f\t\n'
        for item in sort_list:
            if self.is_bool:
                item[2] = str(item[2])
                item[3] = str(item[3])
            line = item_template % tuple(item)
            line = line.replace("nan", "-").replace("inf", "-")
            top_n_list.append(line)
        return top_n_list


class DetailWriter:
    """
    The class for detail writer
    """

    def __init__(self: any, output_path: str, detail_info: DetailInfo) -> None:
        self.output_path = output_path
        self.progress = None
        self.detail_info = detail_info
        self.min_max_value = MinMaxValue()
        self.top_n = TopN()
        self._total_num = 0
        self.dump_file_name = ""

    @staticmethod
    def _padding_shape_to_4d(dim: tuple) -> list:
        left_count = max(0, ConstManager.FOUR_DIMS_LENGTH) - len(dim)
        return list(dim) + left_count * [1]

    @staticmethod
    def _write_one_detail(index: int, file_stream: any, *data_values: any, is_bool: bool = False) -> None:
        cur_dim = data_values[0]
        my_output_value = data_values[1]
        ground_truth_value = data_values[2]
        absolute_error = data_values[3]
        relative_error = data_values[4]
        # add index, n,c,h,w or id, my output value, ground truth value, absolute error, relative error to line
        if not is_bool:
            if np.isnan(relative_error) or np.isinf(relative_error):
                line = '%d,%s,%.6f\t,%.6f\t,%.6f\t,-\n' \
                       % (index, " ".join((str(x) for x in cur_dim)), my_output_value,
                          ground_truth_value, absolute_error)
            else:
                line = '%d,%s,%.6f\t,%.6f\t,%.6f\t,%.6f\t\n' \
                       % (index, " ".join((str(x) for x in cur_dim)), my_output_value,
                          ground_truth_value, absolute_error, relative_error)
        else:
            line = '%d,%s,%s\t,%s\t,-\t,-\n' \
                   % (index, " ".join((str(x) for x in cur_dim)), my_output_value,
                      ground_truth_value)
        # write line to file
        file_stream.write(line)

    @staticmethod
    def _replace_inf_and_nan(nd_array: any) -> any:
        if np.isinf(nd_array).any():
            inf_index = np.isinf(nd_array)
            nd_array[inf_index] = np.nan
            return np.nan_to_num(nd_array)
        return np.nan_to_num(nd_array)

    @staticmethod
    def _calculate_str_length(headers: list, data: list) -> list:
        max_header_list = (len(header.strip()) for header in headers)
        max_column_list = list(max_header_list)
        for _data in data:
            max_data_list = (len(str(_element.strip())) for _element in _data)
            new_max_column_list = []
            for _element in zip(max_data_list, max_column_list):
                new_max_column_list.append(max(_element))
            max_column_list = new_max_column_list
        return max_column_list

    @staticmethod
    def _show_top_n_data(data: list, max_column_list: list) -> None:
        for _data in data:
            for index, _element in enumerate(_data):
                end = '\n' if index == len(_data) - 1 else '\t'
                print(str(_element.strip()).ljust(max_column_list[index], ' '), end=end)

    @staticmethod
    def _show_top_n_header(headers: list, max_column_list: list) -> None:
        for index, header in enumerate(headers):
            end = '\n' if index == len(headers) - 1 else '\t'
            print(str(header.strip()).ljust(max_column_list[index], ' '), end=end)

    def delete_old_detail_result_files(self: any) -> None:
        """
        Delete old detail result files
        """
        if not os.path.exists(self.output_path):
            return
        # delete old result
        for item in os.listdir(self.output_path):
            if item == ConstManager.SIMPLE_OP_MAPPING_FILE_NAME:
                continue
            if self._match_detail_file_name(item):
                path = os.path.join(self.output_path, item)
                if os.path.exists(path):
                    os.remove(path)
        self._delete_too_long_file()

    def write(self: any, dim: tuple, my_output_data: any, ground_truth_data: any, dump_file_name: str = "") -> None:
        """
        Write detail info file
        :param dim: the shape
        :param my_output_data: my output data
        :param ground_truth_data: ground truth data
        :param dump_file_name: dump file name
        """
        self.progress = Progress(len(my_output_data))
        self.dump_file_name = dump_file_name
        old_file_name = '%s.npy' % (self.detail_info.tensor_id.get_file_prefix())
        new_file_name = self._handle_too_long_file_name(old_file_name, ConstManager.NPY_SUFFIX)
        new_file_path = os.path.join(self.output_path, new_file_name)
        FileUtils.save_array_to_file(new_file_path, my_output_data.reshape(dim), np_save=True)
        log.print_write_result_info("%s after shape conversion" % self.detail_info.tensor_id.get_tensor_id(),
                                    new_file_name)
        # check whether the types of my_output_data and ground_truth_data are Boolean.
        self._check_dtype_is_bool(my_output_data, ground_truth_data)
        # transform dim shape to list
        dim_list = self._transform_dim_list(dim)
        # use numpy to calculate the summaries
        res_generator = self._array_calculate(my_output_data, ground_truth_data, dim_list, self.top_n.is_bool)
        # write summary file
        self._write_detail_summary_file(self.top_n.is_bool)
        # write each value to file
        self._write_detail_result_multi_proc(res_generator)
        # finish and print progress
        self.progress.update_and_print_progress(ConstManager.MAX_PROGRESS)
        # write and print top_n
        self._handle_top_n()
        log.print_write_result_info('details result', self.output_path)

    def _write_detail_result_multi_proc(self: any, res_generator: any) -> None:
        if self.detail_info.ignore_result:
            return
        # prepare task pool
        cpu_count = int((multiprocessing.cpu_count() + 1) / 2)
        pool = multiprocessing.Pool(cpu_count)

        # allocate all tasks evenly by multi-processing number
        task_list = []
        for _ in range(cpu_count):
            task_list.append([])
        for i, group in enumerate(res_generator):
            task_list[i % cpu_count].append(group)

        # make a listen process to listen the print progress
        comm_queue = multiprocessing.Manager().Queue()
        listen_proc = multiprocessing.Process(target=self._listen_and_print,
                                              args=(comm_queue,))
        listen_proc.start()

        # apply tasks
        for task in task_list:
            pool.apply_async(self._multi_process_group_write,
                             args=(task, comm_queue,))
        # wait till done
        pool.close()
        pool.join()
        listen_proc.terminate()

    def _listen_and_print(self: any, queue: multiprocessing.Queue) -> None:
        while True:
            add_num = queue.get(True)
            self.progress.update_progress(add_num)
            if not self.progress.is_done():
                self.progress.print_progress()

    def _separate_group_generator(self: any, dim_list: list, data_group: tuple, is_bool: bool = False) -> tuple:
        # generate the slice data of input lists
        i = 0
        length = len(dim_list)
        my_output_data = data_group[0]
        ground_truth_data = data_group[1]
        absolute_error = data_group[2]
        relative_error = data_group[3]
        max_line = self.detail_info.max_line
        while (i + 1) * max_line < length:
            left = i * max_line
            right = (i + 1) * max_line
            s = slice(left, right)
            yield i, dim_list[s], my_output_data[s], ground_truth_data[s], absolute_error[s], relative_error[s], is_bool
            i = i + 1

        s = slice(i * max_line, length)
        yield i, dim_list[s], my_output_data[s], ground_truth_data[s], absolute_error[s], relative_error[s], is_bool

    def _group_write_exec(self: any, task_group: list, queue: multiprocessing.Queue) -> None:
        for arg_group in task_group:
            task_index = arg_group[0]
            dim_list = arg_group[1]
            my_output_data = arg_group[2]
            ground_truth_data = arg_group[3]
            absolute_error = arg_group[4]
            relative_error = arg_group[5]
            # make header, should do once
            file_path = self._make_detail_output_file(task_index * self.detail_info.max_line)
            # create file IO pointer
            detail_output_file = os.fdopen(os.open(file_path, ConstManager.WRITE_FLAGS, ConstManager.WRITE_MODES),
                                           'w')
            detail_output_file.write(self.detail_info.make_detail_header())
        
            # write detail results one by one
            for offset, dim in enumerate(dim_list):
                index = task_index * self.detail_info.max_line + offset
                self._write_one_detail(index, detail_output_file, dim, my_output_data[offset],
                                       ground_truth_data[offset], absolute_error[offset], relative_error[offset],
                                       is_bool=arg_group[6])
        
            detail_output_file.close()
        
            # add the write count to the shared queue
            queue.put(len(dim_list))

    def _multi_process_group_write(self: any, task_group: list, queue: multiprocessing.Queue) -> None:
        try:
            self._group_write_exec(task_group, queue)
        except Exception as error:
            log.print_error_log('Failed to write detail info in subprocess. %s' % error)

    def _transform_dim_list(self: any, dim: tuple) -> list:
        if self.detail_info.detail_format != '()':
            # padding shape to 4d
            dims = self._padding_shape_to_4d(dim)
            args = (range(x) for x in dims)
            # write each detail value
            cur_dim_list = list(itertools.product(*args))
        else:
            cur_dim_list = []
            for cur_dim0 in range(dim[0]):
                cur_dim_list.append([cur_dim0])

        return cur_dim_list

    def _match_detail_file_name(self: any, file_name: str) -> bool:
        name = self.detail_info.tensor_id.get_file_prefix()
        return file_name == ("%s_summary.txt" % name) or (file_name.startswith("%s_" % name) and (
                file_name.endswith(ConstManager.CSV_SUFFIX) or file_name.endswith(ConstManager.NPY_SUFFIX)))

    def _make_detail_output_file(self: any, file_index: int = 0) -> str:
        # make file name based on the index，then return the path
        old_file_name = "%s_%d.csv" % (self.detail_info.tensor_id.get_file_prefix(), file_index)
        new_file_name = self._handle_too_long_file_name(old_file_name, ConstManager.CSV_SUFFIX, file_index)
        file_path = os.path.join(self.output_path, new_file_name)

        return file_path

    def _array_calculate(self: any, my_output_data: any, ground_truth_data: any, dim_list: list,
                         is_bool: bool = False) -> any:
        self._total_num = len(my_output_data)
        absolute_error, relative_error = self._cal_err(my_output_data, ground_truth_data, is_bool)
        # find the index of top n error values
        top_n = self.detail_info.top_n
        if len(absolute_error) < top_n:
            top_n = len(absolute_error) - 1
        absolute_top_n_index, relative_top_n_index = self._get_error_top_n_index(absolute_error, relative_error, top_n,
                                                                                 is_bool)
        # record top n lines in calculator
        self._set_top_n_list("absolute", absolute_top_n_index, dim_list,
                             (my_output_data[absolute_top_n_index], ground_truth_data[absolute_top_n_index],
                              absolute_error[absolute_top_n_index], relative_error[absolute_top_n_index]))
        self._set_top_n_list("relative", relative_top_n_index, dim_list,
                             (my_output_data[relative_top_n_index], ground_truth_data[relative_top_n_index],
                              absolute_error[relative_top_n_index], relative_error[relative_top_n_index]))

        # return the generator of compute result, to save memory
        return self._separate_group_generator(dim_list,
                                              (my_output_data, ground_truth_data, absolute_error, relative_error),
                                              is_bool)

    def _cal_err(self: any, my_output_data: any, ground_truth_data: any, is_bool: bool) -> tuple:
        if not is_bool:
            # use numpy to calculate summary data quickly
            absolute_error = np.abs(my_output_data - ground_truth_data)
            abs_ground_data = np.abs(ground_truth_data)
            relative_error = np.true_divide(absolute_error, abs_ground_data)
        else:
            absolute_error = np.array([np.nan] * self._total_num)
            relative_error = np.array([np.nan] * self._total_num)
        return absolute_error, relative_error

    def _get_error_top_n_index(self: any, absolute_error: any, relative_error: any, top_n: int, is_bool: bool) -> tuple:
        if not is_bool:
            absolute_new_error = self._replace_inf_and_nan(absolute_error)
            relative_new_error = self._replace_inf_and_nan(relative_error)
            absolute_top_n_index = np.argpartition(absolute_new_error, -top_n)[-top_n:].tolist()
            relative_top_n_index = np.argpartition(relative_new_error, -top_n)[-top_n:].tolist()
            if len(absolute_new_error) != 0:
                self.min_max_value.set_min_absolute_error(np.nanmin(absolute_new_error))
                self.min_max_value.set_max_absolute_error(np.nanmax(absolute_new_error))
            if len(relative_new_error) != 0:
                self.min_max_value.set_min_relative_error(np.nanmin(relative_new_error))
                self.min_max_value.set_max_relative_error(np.nanmax(relative_new_error))
        else:
            absolute_top_n_index = list(range(top_n))
            relative_top_n_index = list(range(top_n))

            self.min_max_value.set_min_absolute_error(np.nan)
            self.min_max_value.set_max_absolute_error(np.nan)
            self.min_max_value.set_min_relative_error(np.nan)
            self.min_max_value.set_max_relative_error(np.nan)
        return absolute_top_n_index, relative_top_n_index

    def _set_top_n_list(self: any, error_type: str, index_list: list, dim_list: list, top_n_data_group: tuple) -> None:
        top_n_dims = []
        for i in index_list:
            top_n_dims.append(dim_list[i])
        if error_type == "absolute":
            self.top_n.set_absolute_top_n_list(index_list, top_n_dims, top_n_data_group[0], top_n_data_group[1],
                                               top_n_data_group[2], top_n_data_group[3], error_type)
        else:
            self.top_n.set_relative_top_n_list(index_list, top_n_dims, top_n_data_group[0], top_n_data_group[1],
                                               top_n_data_group[2], top_n_data_group[3], error_type)

    def _check_dtype_is_bool(self: any, my_output_data: any, ground_truth_data: any) -> None:
        is_bool = my_output_data.dtype == np.bool_ and ground_truth_data.dtype == np.bool_
        self.top_n.is_bool = is_bool

    def _write_top_n_file(self: any, top_n_type: str, top_n_list: list) -> None:
        old_file_name = "%s_%s_error_topn.csv" % (self.detail_info.tensor_id.get_file_prefix(), top_n_type)
        top_n_type_suffix = "_%s_error_topn.csv" % top_n_type
        new_file_name = self._handle_too_long_file_name(old_file_name, top_n_type_suffix)
        top_n_path = os.path.join(self.output_path, new_file_name)
        path_check.check_write_path_secure(top_n_path)
        try:
            with os.fdopen(os.open(top_n_path, ConstManager.WRITE_FLAGS, ConstManager.WRITE_MODES), 'w') as top_n_file:
                top_n_file.write(self.detail_info.make_detail_header())
                for line in top_n_list:
                    top_n_file.write(line)
        except (OSError, SystemError, ValueError, TypeError, RuntimeError,
                MemoryError, KeyError, IOError) as error:
            log.print_error_log('Failed to write top n file. %s' % error)
            raise CompareError(CompareError.MSACCUCMP_WRITE_FILE_ERROR) from error

    def _make_show_data(self: any, top_n_list: list) -> (list, list):
        header = self.detail_info.make_detail_header().split(',')
        data = []
        for item in top_n_list:
            data.append(item.split(','))
        return header, data

    def _show_top_n(self: any, top_n_type: str, top_n_list: list) -> None:
        headers, data = self._make_show_data(top_n_list)
        max_column_list = self._calculate_str_length(headers, data)
        log.print_info_log("%s TopN" % top_n_type)
        self._show_top_n_header(headers, max_column_list)
        self._show_top_n_data(data, max_column_list)

    def _handle_top_n(self: any) -> None:
        print("\n")
        absolute_top_n = self.top_n.get_absolute_error_top_n_list()
        self._write_top_n_file('absolute', absolute_top_n)
        self._show_top_n("Absolute Error", absolute_top_n)
        print("\n")
        relative_top_n = self.top_n.get_relative_error_top_n_list()
        self._write_top_n_file('relative', relative_top_n)
        self._show_top_n("Relative Error", relative_top_n)

    def _write_detail_summary_file(self: any, is_bool: bool = False) -> None:
        content = "TotalCount:%d\n%s%s\n" \
                  % (self._total_num, self.detail_info.get_detail_info(), self.min_max_value.get_min_and_max_value())
        if is_bool:
            log.print_warn_log('Boolean data does not support calculate Relative Error or Absolute Error!')
            content = content.replace('nan', '-')
        old_file_name = "%s_summary.txt" % (self.detail_info.tensor_id.get_file_prefix())
        new_file_name = self._handle_too_long_file_name(old_file_name, ConstManager.SUMMARY_TXT_SUFFIX)
        summary_file_path = os.path.join(self.output_path, new_file_name)
        path_check.check_write_path_secure(summary_file_path)
        try:
            with os.fdopen(os.open(summary_file_path, ConstManager.WRITE_FLAGS, ConstManager.WRITE_MODES),
                           'w') as summary_file:
                summary_file.write(content)
        except (OSError, SystemError, ValueError, TypeError, RuntimeError,
                MemoryError, KeyError, IOError) as error:
            log.print_error_log('Failed to write detail summary file. %s' % error)
            raise CompareError(CompareError.MSACCUCMP_WRITE_FILE_ERROR) from error

    def _handle_too_long_file_name(self: any, file_name: str, suffix: str, file_index: int = 0) -> str:
        if len(file_name) >= ConstManager.LINUX_FILE_NAME_MAX_LEN:
            tensor_type_index = self.detail_info.tensor_id.get_tensor_type_index()
            new_file_name = "%s%s%s" % (self.dump_file_name.replace('/', '_').replace('.', '_'),
                                        tensor_type_index, suffix)
            if ConstManager.CSV_SUFFIX == suffix:
                new_file_name = "%s%s_%s%s" % (self.dump_file_name.replace('/', '_').replace('.', '_'),
                                               tensor_type_index, file_index, suffix)
            if len(new_file_name) >= ConstManager.LINUX_FILE_NAME_MAX_LEN:
                value = ''.join(str(uuid.uuid3(uuid.NAMESPACE_DNS, file_name)).split('-'))
                new_file_name = "%s%s" % (value, suffix)
            self._save_op_mapping_file(file_name, new_file_name)
            return new_file_name
        return file_name

    def _save_op_mapping_file(self: any, old_file_name: str, new_file_name: str) -> None:
        single_op_mapping_path = os.path.join(self.output_path, ConstManager.SIMPLE_OP_MAPPING_FILE_NAME)
        try:
            with os.fdopen(os.open(single_op_mapping_path, ConstManager.WRITE_FLAGS, ConstManager.WRITE_MODES),
                           'a+') as mapping_file:
                content = "%s,%s\n" % (new_file_name, old_file_name)
                mapping_file.write(content)
        except (OSError, SystemError, ValueError, TypeError, RuntimeError,
                MemoryError, KeyError, IOError) as error:
            log.print_error_log('Failed to save mapping file. %s' % error)
            raise CompareError(CompareError.MSACCUCMP_WRITE_FILE_ERROR) from error

    def _delete_file_by_mapping(self: any, hash_to_file_name_map: dict) -> list:
        delete_keys = []
        # delete old result
        for key, value in hash_to_file_name_map.items():
            if self._match_detail_file_name(value):
                path = os.path.join(self.output_path, key)
                if os.path.exists(path):
                    os.remove(path)
                delete_keys.append(key)
        return delete_keys

    def _delete_too_long_file(self: any) -> None:
        mapping_file_path = os.path.join(self.output_path, ConstManager.SIMPLE_OP_MAPPING_FILE_NAME)
        if not os.path.exists(mapping_file_path):
            return
        hash_to_file_name_map = mapping.read_mapping_file(mapping_file_path)
        delete_keys = self._delete_file_by_mapping(hash_to_file_name_map)
        for key in delete_keys:
            hash_to_file_name_map.pop(key)
        if os.path.exists(mapping_file_path):
            os.remove(mapping_file_path)
        # save other compare result records.
        if hash_to_file_name_map:
            with os.fdopen(os.open(mapping_file_path, ConstManager.WRITE_FLAGS, ConstManager.WRITE_MODES),
                           "w") as output_file:
                for key, value in hash_to_file_name_map.items():
                    content = "%s,%s\n" % (key, value)
                    output_file.write(content)



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
import shutil

import pandas as pd

from msprobe.core.common.file_utils import save_excel, split_zip_file_path, \
    create_directory, extract_zip
from msprobe.core.common.framework_adapter import FmkAdp
from msprobe.core.config_check.checkers.base_checker import PackInput
from msprobe.core.config_check.utils.utils import config_checking_print
from msprobe.core.common.const import Const


class ConfigChecker:
    checkers = {}
    pre_forward_fun_list = []
    result_filename = "result.xlsx"
    result_header = ["filename", "pass_check"]
    step = 0

    def __init__(self, model=None, shell_path=None, output_zip_path="./config_check_pack.zip", fmk="pytorch"):
        if model:
            FmkAdp.set_fmk(fmk)
        self.pack_input = PackInput(output_zip_path, model, shell_path)
        file_path, file_name = split_zip_file_path(self.pack_input.output_zip_path)
        if not os.path.exists(file_path):
            create_directory(file_path)
        self.pack()

    @staticmethod
    def compare(bench_zip_path, cmp_zip_path, output_path):
        create_directory(output_path)
        bench_dir = os.path.join(output_path, "bench")
        cmp_dir = os.path.join(output_path, "cmp")
        extract_zip(bench_zip_path, bench_dir)
        config_checking_print(f"extract zip file {bench_zip_path} to {bench_dir}")
        extract_zip(cmp_zip_path, cmp_dir)
        config_checking_print(f"extract zip file {cmp_zip_path} to {cmp_dir}")

        result = []
        summary_result = []
        for checker in ConfigChecker.checkers.values():
            checker_name, pass_check, df = checker.compare_ex(bench_dir, cmp_dir, output_path)
            if checker_name:
                summary_result.append([checker_name, pass_check])
            if df is not None:
                result.append((df, checker_name))
        summary_result_df = pd.DataFrame(summary_result, columns=ConfigChecker.result_header)
        result.insert(0, (summary_result_df, "summary"))
        save_excel(os.path.join(output_path, ConfigChecker.result_filename), result)
        config_checking_print(f"config checking result save to {os.path.realpath(output_path)}")

    @staticmethod
    def apply_patches(fmk=Const.PT_FRAMEWORK):
        for checker in ConfigChecker.checkers.values():
            checker.apply_patches(fmk)

    def pack(self):
        config_checking_print(f"pack result zip path {os.path.realpath(self.pack_input.output_zip_path)}")

        def hook(model, args, kwargs):
            for collect_func in self.pre_forward_fun_list:
                collect_func(model, args, kwargs, ConfigChecker.step)
            ConfigChecker.step += 1

        if self.pack_input.model:
            FmkAdp.register_forward_pre_hook(self.pack_input.model, hook, with_kwargs=True)
        for checker in ConfigChecker.checkers.values():
            if checker.input_needed and not getattr(self.pack_input, checker.input_needed):
                continue
            if FmkAdp.is_initialized() and FmkAdp.get_rank() != 0 and not checker.multi_rank:
                continue
            checker.pack(self.pack_input)


def register_checker_item(key, cls=None):
    if cls is None:
        # 无参数时，返回装饰器函数
        return lambda cls: register_checker_item(key, cls)
    ConfigChecker.checkers[key] = cls
    return cls


def register_pre_forward_fun_list(func):
    ConfigChecker.pre_forward_fun_list.append(func)



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from msprobe.core.compare.acc_compare import Comparator, ModeConfig, MappingConfig, setup_comparison
from msprobe.pytorch.compare.utils import read_pt_data


def read_real_data(npu_dir, npu_data_name, bench_dir, bench_data_name, _) -> tuple:
    n_value = read_pt_data(npu_dir, npu_data_name)
    b_value = read_pt_data(bench_dir, bench_data_name)
    return n_value, b_value


def pt_compare(input_param, output_path, **kwargs):
    config = setup_comparison(input_param, output_path, **kwargs)

    config_dict = {
        'stack_mode': config.stack_mode,
        'fuzzy_match': config.fuzzy_match,
        'dump_mode': config.dump_mode,
        'first_diff_analyze': config.first_diff_analyze,
        'compared_file_type': config.compared_file_type
    }
    mode_config = ModeConfig(**config_dict)
    mapping_config = MappingConfig(cell_mapping=config.cell_mapping, data_mapping=config.data_mapping)
    pt_comparator = Comparator(read_real_data, mode_config, mapping_config)
    pt_comparator.compare_core(input_param, output_path, suffix=config.suffix)



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
import zlib
from concurrent.futures import ThreadPoolExecutor
import inspect

import mindspore as ms
from mindspore import mint, ops, hal
from mindspore.mint import distributed
from mindspore._c_expression.typing import Number
import numpy as np

from msprobe.core.common.const import Const
from msprobe.core.dump.data_dump.data_processor.base import (BaseDataProcessor, TensorStatInfo)
from msprobe.mindspore.common.utils import convert_bf16_to_fp32, save_tensor_as_npy
from msprobe.mindspore.common.log import logger
from msprobe.mindspore.dump.dump_processor.hook_cell.api_register import get_api_register
from msprobe.mindspore.common.utils import is_mindtorch, cast_to_float_if_fp8

has_adump = True
try:
    from msprobe.lib import _msprobe_c
except ImportError:
    has_adump = False

if is_mindtorch():
    from torch import distributed as dist


class MindsporeDataProcessor(BaseDataProcessor):
    if is_mindtorch():
        mindspore_special_type = tuple([ms.Tensor, Number, distributed.P2POp, dist.ProcessGroup])
    else:
        mindspore_special_type = tuple([ms.Tensor, Number, distributed.P2POp])

    def __init__(self, config, data_writer):
        super().__init__(config, data_writer)
        self.mindspore_object_key = {
            "dtype": self.analyze_dtype_in_kwargs
        }
        self._async_dump_cache = {}
        self.api_register = get_api_register()
        self._crc_executor = ThreadPoolExecutor(max_workers=os.cpu_count() // 2)

    @staticmethod
    def compute_crc32_bytes(tensor_bytes):
        return f"{zlib.crc32(tensor_bytes):08x}"

    @staticmethod
    def get_md5_for_tensor(x):
        x = convert_bf16_to_fp32(x)
        tensor_bytes = x.asnumpy().tobytes()
        crc32_hash = zlib.crc32(tensor_bytes)
        return f"{crc32_hash:08x}"

    @staticmethod
    def analyze_dtype_in_kwargs(element):
        return {"type": "mindspore.dtype", "value": str(element)}

    @staticmethod
    def is_hookable_element(element):
        return hasattr(element, "register_hook") and callable(element.register_hook)

    @staticmethod
    def process_group_hash(arg):
        group_ranks = distributed.get_process_group_ranks(arg)
        group_ranks_hash = zlib.crc32(str(group_ranks).encode('utf-8'))
        return f"{group_ranks_hash:08x}"

    @staticmethod
    def _analyze_process_group(arg):
        group_info = {"type": "mindspore.ProcessGroup"}
        try:
            group_ranks = dist.get_process_group_ranks(arg)
            group_info.update({"group_ranks": group_ranks})
            group_ranks_hash = zlib.crc32(str(group_ranks).encode('utf-8'))
            group_id = f"{group_ranks_hash:08x}"
            group_info.update({"group_id": group_id})
        except Exception as e:
            logger.warning(f"Failed to get process group ranks info with error info: {e}.")
        return group_info

    @classmethod
    def get_special_types(cls):
        return super().get_special_types() + cls.mindspore_special_type

    def dump_async_data(self):
        for file_path, tensor in self._async_dump_cache.items():
            save_tensor_as_npy(tensor, file_path)
        self._async_dump_cache.clear()

    def get_stat_info(self, data):
        self.api_register.restore_inner_used_api()
        tensor_stat = TensorStatInfo()
        if data.numel() == 0:
            pass
        elif data.dtype == ms.bool_:
            if self.config.async_dump:
                tensor_stat.max = mint.any(data)
                tensor_stat.min = mint.all(data)
            else:
                data_np = data.asnumpy()
                tensor_stat.max = np.max(data_np).item()
                tensor_stat.min = np.min(data_np).item()
        elif not data.shape:
            tensor_stat.max = tensor_stat.min = tensor_stat.mean = tensor_stat.norm = data.copy()
        elif data.dtype == ms.complex64 or data.dtype == ms.complex128:
            if self.config.async_dump:
                logger.warning("Async dump do not support complex data!")
            else:
                data_abs = np.abs(data.asnumpy())
                tensor_stat.max = np.max(data_abs).item()
                tensor_stat.min = np.min(data_abs).item()
                tensor_stat.mean = np.mean(data_abs).item()
                tensor_stat.norm = np.linalg.norm(data_abs).item()
        else:
            if self.config.precision == Const.DUMP_PRECISION_HIGH or not ops.is_floating_point(
                    data) or data.dtype == ms.float64:
                data = data.to(ms.float32)
            get_norm_value = mint.norm if hasattr(mint, "norm") else ops.norm
            tensor_stat.max = mint.max(data)
            tensor_stat.min = mint.min(data)
            tensor_stat.mean = mint.mean(data)
            tensor_stat.norm = get_norm_value(data)
        self.api_register.register_inner_used_api()
        return tensor_stat

    def analyze_single_element(self, element, suffix_stack):
        if suffix_stack and suffix_stack[-1] in self.mindspore_object_key:
            return self.mindspore_object_key[suffix_stack[-1]](element)

        suffix_str = Const.SEP.join(str(s) for s in suffix_stack)
        type_analyzer = [
            (MindsporeDataProcessor.builtin_type, self._analyze_builtin),
            (ms.Tensor, lambda e: self._analyze_tensor(e, suffix_str)),
            (Number, self.analyze_dtype_in_kwargs),
            (MindsporeDataProcessor.np_type[:-1], self._analyze_numpy),
            (np.ndarray, lambda e: self._analyze_ndarray(e, suffix_str)),
            (distributed.P2POp, lambda e: self._analyze_p2pop(e, suffix_str))
        ]
        if is_mindtorch():
            type_analyzer.append((dist.ProcessGroup, self._analyze_process_group))
        for type_key, analyze_fn in type_analyzer:
            if isinstance(element, type_key):
                return analyze_fn(element)
        return {}

    def _analyze_p2pop(self, arg, suffix):
        p2pop_info = {"class_type": "mindspore.mint.distributed.P2POp"}
        try:
            tensor_info = self._analyze_tensor(arg.tensor, suffix)
            p2pop_info.update({"tensor": tensor_info})
            p2pop_info.update({"op": arg.op})
            p2pop_info.update({"peer": arg.peer})
            p2pop_info.update({"tag": arg.tag})
            group_id = self.process_group_hash(arg.group) if arg.group else None
            p2pop_info.update({"group_id": group_id})
        except Exception as e:
            logger.warning(f"Failed to parse the P2POp content with error info: {e}.")
        return p2pop_info

    def _analyze_tensor(self, tensor, suffix):
        dtype = str(tensor.dtype)
        tensor = cast_to_float_if_fp8(tensor)
        tensor_stat = self.get_stat_info(tensor)
        tensor_json = {
            'type': 'mindspore.Tensor',
            'dtype': dtype,
            'shape': tensor.shape
        }
        if hasattr(tensor, "layout") and tensor.layout is not None:
            layout_dic = {}
            if hasattr(tensor.layout, "device_matrix") and tensor.layout.device_matrix is not None:
                layout_dic['Device Matrix'] = tensor.layout.device_matrix
            if hasattr(tensor.layout, "alias_name") and tensor.layout.alias_name is not None:
                layout_dic['Alias Name'] = tensor.layout.alias_name
                interleaved = "Yes" if ("interleaved_parallel" in tensor.layout.alias_name) else "No"
                layout_dic['Interleaved'] = interleaved
            if hasattr(tensor.layout, "partial") and tensor.layout.partial is not None:
                layout_dic['Partial'] = tensor.layout.partial
            if hasattr(tensor.layout, "tensor_map") and tensor.layout.tensor_map is not None:
                layout_dic['Tensor Map'] = tensor.layout.tensor_map
            if hasattr(tensor.layout, "rank_list") and tensor.layout.rank_list is not None:
                layout_dic['Rank List'] = tensor.layout.rank_list
            tensor_json['layout'] = layout_dic
        if hasattr(tensor, "hsdp_effective_shard_size") and tensor.hsdp_effective_shard_size is not None:
            tensor_json['hsdp_shard_size'] = tensor.hsdp_effective_shard_size

        # 将统计值存入全局 buffer，并返回占位索引
        stat_values = [
            tensor_stat.max,
            tensor_stat.min,
            tensor_stat.mean,
            tensor_stat.norm
        ]

        placeholder_index = self.data_writer.append_stat_to_buffer(stat_values)

        tensor_json.update({Const.TENSOR_STAT_INDEX: placeholder_index})

        if self.config.summary_mode == Const.MD5 and not self.config.async_dump:
            tensor = convert_bf16_to_fp32(tensor)
            # 拷贝并搬到 CPU
            tensor_bytes = tensor.asnumpy()

            future = self._crc_executor.submit(
                MindsporeDataProcessor.compute_crc32_bytes,
                tensor_bytes
            )

            crc_placeholder = self.data_writer.append_crc32_to_buffer(future)
            tensor_json[Const.MD5_INDEX] = crc_placeholder

        return tensor_json

    def _analyze_and_save_tensor(self, tensor, suffix):
        dump_data_name, file_path = self.get_save_file_path(suffix)
        single_arg = MindsporeDataProcessor._analyze_tensor(self, tensor, suffix)
        single_arg.update({"data_name": dump_data_name})
        tensor = cast_to_float_if_fp8(tensor)
        if self.config.async_dump:
            self._async_dump_cache[file_path] = tensor.copy()
        else:
            save_tensor_as_npy(tensor, file_path)
        return single_arg


class StatisticsDataProcessor(MindsporeDataProcessor):
    def _analyze_tensor(self, tensor, suffix):
        if any(item in self.current_api_or_module_name for item in self.config.tensor_list):
            return self._analyze_and_save_tensor(tensor, suffix)
        else:
            return super()._analyze_tensor(tensor, suffix)

    def _analyze_ndarray(self, ndarray, suffix):
        if any(item in self.current_api_or_module_name for item in self.config.tensor_list):
            return self._analyze_and_save_ndarray(ndarray, suffix)
        else:
            return super()._analyze_ndarray(ndarray, suffix)


class TensorDataProcessor(MindsporeDataProcessor):
    def _analyze_tensor(self, tensor, suffix):
        return self._analyze_and_save_tensor(tensor, suffix)

    def _analyze_ndarray(self, ndarray, suffix):
        return self._analyze_and_save_ndarray(ndarray, suffix)


class KernelDumpDataProcessor(MindsporeDataProcessor):
    def __init__(self, config, data_writer):
        super().__init__(config, data_writer)
        self.enable_kernel_dump = True

    @staticmethod
    def start_kernel_dump(config_path):
        hal.synchronize()
        _msprobe_c.init_dump()
        _msprobe_c.set_dump(config_path)
        hal.synchronize()

    @staticmethod
    def stop_kernel_dump():
        hal.synchronize()
        _msprobe_c.finalize_dump()
        hal.synchronize()

    @staticmethod
    def _print_unsupported_log(api_name):
        logger.warning(f"The kernel dump does not support the {api_name} API.")

    def analyze_forward_input(self, name, module, module_input_output):
        if not self.enable_kernel_dump:
            return
        if not has_adump:
            logger.warning("The current msprobe package does not compile adump, and kernel dump cannot be used.")
            self.enable_kernel_dump = False
            return
        self.start_kernel_dump(self.config.kernel_config_path)

    def analyze_forward_output(self, name, module, module_input_output):
        if not self.enable_kernel_dump:
            return
        self.enable_kernel_dump = False
        self.stop_kernel_dump()
        logger.info(f"The kernel data of {name} is dumped successfully.")

    def analyze_backward_input(self, name, module, module_input_output):
        if not self.enable_kernel_dump:
            return
        if not has_adump:
            logger.warning("The current msprobe package does not compile adump, and kernel dump cannot be used.")
            self.enable_kernel_dump = False
            return
        self.start_kernel_dump(self.config.kernel_config_path)

    def analyze_backward(self, name, module, module_input_output):
        if not self.enable_kernel_dump:
            return
        self.enable_kernel_dump = False
        self.stop_kernel_dump()
        logger.info(f"The kernel data of {name} is dumped successfully.")

    def reset_status(self):
        self.enable_kernel_dump = True



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
import threading

from mindspore import ops
from mindspore.common.tensor import Tensor
from msprobe.core.common.utils import Const, DumpException, ThreadSafe
from msprobe.core.dump.data_dump.data_processor.base import (
    ModuleBackwardInputs,
    ModuleBackwardOutputs,
    ModuleForwardInputsOutputs
)
from msprobe.core.dump.hook_manager import BaseHookManager
from msprobe.mindspore.common.log import logger


class PrimitiveHookService:
    def __init__(self, service_instance):
        self.primitive_counters = {}
        self.service_instance = service_instance

    def wrap_primitive(self, origin_func, primitive_name):
        """
        包装原始的 primitive 函数，添加输入和输出的 hook 以捕获前向和反向数据。

        Args:
            origin_func (callable): 原始 的 primitive 函数。
            primitive_name (str): 原始的 primitive 名称。

        Returns:
            callable: 包装后的 primitive 函数。
        """

        def create_backward_hook(captured_grads, num_tensors, updated_primitive_name, hook_type):
            """
            创建反向 hook 函数，用于捕获梯度。

            Args:
                captured_grads (list): 用于保存捕获的梯度。
                num_tensors (int): 张量数量。
                updated_primitive_name (str): 更新后的 primitive 名称。
                hook_type (str): hook 类型 (输入/输出)。

            Returns:
                callable: 反向 hook 函数。
            """

            @ThreadSafe.synchronized
            def backward_hook(grad):
                tid = threading.get_ident()
                BaseHookManager.inner_switch[tid] = True

                captured_grads.extend(grad)
                backward_primitive_name = f"{updated_primitive_name}{Const.SEP}{Const.BACKWARD}"
                try:
                    if hook_type == Const.INPUT:
                        self.service_instance.data_collector.update_api_or_module_name(backward_primitive_name)
                        new_module_input_output = ModuleBackwardOutputs(grad_output=tuple(captured_grads))
                        self.service_instance.data_collector.backward_output_data_collect(
                            backward_primitive_name, self, os.getpid(), new_module_input_output
                        )
                    elif hook_type == Const.OUTPUT:
                        self.service_instance.data_collector.update_api_or_module_name(backward_primitive_name)
                        new_module_input_output = ModuleBackwardInputs(grad_input=tuple(captured_grads))
                        self.service_instance.data_collector.backward_input_data_collect(
                            backward_primitive_name, self, os.getpid(), new_module_input_output
                        )

                except Exception as exception:
                    logger.error(f"This is a primitive op {hook_type}_backward dump error: {exception}, "
                                 f"updated_primitive_name: {updated_primitive_name}")
                    raise DumpException(DumpException.BACKWARD_DATA_COLLECTION_ERROR) from exception
                BaseHookManager.inner_switch[tid] = False

            return backward_hook

        def hook_primitive_inputs(args, captured_grads_input, updated_primitive_name):
            """
            针对前向输入添加 hook。

            Args:
                args (tuple): primitive 输入参数。
                captured_grads_input (list): 捕获的输入梯度。
                updated_primitive_name (str): 更新后的 primitive 名称。

            Returns:
                list: 添加了 hook 的输入。
            """
            hooked_inputs = []
            num_tensors = sum(isinstance(arg, Tensor) for arg in args)
            input_backward_hook = create_backward_hook(captured_grads_input, num_tensors, updated_primitive_name,
                                                       Const.INPUT)
            for arg in args:
                if isinstance(arg, Tensor):
                    arg_hooked = ops.HookBackward(input_backward_hook)(arg)
                    hooked_inputs.append(arg_hooked)
                else:
                    hooked_inputs.append(arg)
            return tuple(hooked_inputs)

        def hook_primitive_outputs(out, captured_grads_output, updated_primitive_name):
            """
            针对前向输出添加 hook。

            Args:
                out (Tensor/tuple): primitive 输出。
                captured_grads_output (list): 捕获的输出梯度。
                updated_primitive_name (str): 更新后的 primitive 名称。

            Returns:
                Tensor/tuple: 添加了 hook 的输出。
            """
            if isinstance(out, tuple):
                num_output_tensors = sum(isinstance(tensor, Tensor) for tensor in out)
            else:
                num_output_tensors = 1
            output_backward_hook = create_backward_hook(captured_grads_output, num_output_tensors,
                                                        updated_primitive_name, Const.OUTPUT)

            if isinstance(out, Tensor):
                return ops.HookBackward(output_backward_hook)(out)
            elif isinstance(out, tuple):
                hooked_outputs = []
                for tensor in out:
                    if isinstance(tensor, Tensor):
                        hooked_outputs.append(ops.HookBackward(output_backward_hook)(tensor))
                    else:
                        hooked_outputs.append(tensor)
                return tuple(hooked_outputs)
            return out

        @ThreadSafe.synchronized
        def pre_forward_hook(primitive_name, primitive_instance, args, kwargs):
            tid = threading.get_ident()
            BaseHookManager.inner_switch[tid] = True

            self.service_instance.data_collector.update_api_or_module_name(primitive_name)
            module_input_output = ModuleForwardInputsOutputs(args=args, kwargs=kwargs, output=None)
            try:
                self.service_instance.data_collector.forward_input_data_collect(
                    primitive_name,
                    primitive_instance,
                    os.getpid(),
                    module_input_output
                )
            except Exception as exception:
                logger.error(f"This is a primitive op dump error during forward input data collection: {exception}, "
                             f"primitive_name: {primitive_name}")
                raise DumpException(DumpException.FORWARD_DATA_COLLECTION_ERROR) from exception
            BaseHookManager.inner_switch[tid] = False

        @ThreadSafe.synchronized
        def post_forward_hook(primitive_name, primitive_instance, args, kwargs, output):
            tid = threading.get_ident()
            BaseHookManager.inner_switch[tid] = True

            self.service_instance.data_collector.update_api_or_module_name(primitive_name)
            module_input_output = ModuleForwardInputsOutputs(args=args, kwargs=kwargs, output=output)
            try:
                self.service_instance.data_collector.forward_output_data_collect(
                    primitive_name,
                    primitive_instance,
                    os.getpid(),
                    module_input_output
                )
            except Exception as exception:
                logger.error(f"This is a primitive op dump error during forward output data collection: {exception}, "
                             f"primitive_name: {primitive_name}")
                raise DumpException(DumpException.FORWARD_DATA_COLLECTION_ERROR) from exception
            BaseHookManager.inner_switch[tid] = False

        def wrapped_primitive_call(instance_self, *args, **kwargs):
            """
            包装后的 primitive 调用函数，添加输入和输出的 hook。

            Args:
                instance_self (object): primitive 的实例。
                *args: primitive 输入参数。
                **kwargs: primitive 关键字参数。

            Returns:
                Tensor/tuple: primitive 的返回值。
            """
            self.update_primitive_counters(primitive_name)
            current_count = self.primitive_counters.get(primitive_name, 0)
            updated_primitive_name = f"{Const.PRIMITIVE_PREFIX}{Const.SEP}{primitive_name}{Const.SEP}{current_count}"

            tid = threading.get_ident()
            if not self.service_instance.primitive_switch or BaseHookManager.inner_switch[tid]:
                return origin_func(*args, **kwargs)

            captured_grads_input, captured_grads_output = [], []

            try:
                hooked_inputs = hook_primitive_inputs(args, captured_grads_input, updated_primitive_name)
            except Exception as exception:
                logger.error(f"This is a primitive op dump error during input hooking: {exception}, "
                             f"primitive_name: {primitive_name}")
                raise DumpException(DumpException.INPUT_HOOK_ERROR) from exception

            forward_primitive_name = f"{updated_primitive_name}{Const.SEP}{Const.FORWARD}"
            pre_forward_hook(forward_primitive_name, instance_self, hooked_inputs, kwargs)
            try:
                out = origin_func(*hooked_inputs, **kwargs)
            except Exception as exception:
                logger.error(f"This is a primitive op dump error during function call: {exception}, "
                             f"primitive_name: {primitive_name}")
                raise DumpException(DumpException.FUNCTION_CALL_ERROR) from exception
            post_forward_hook(forward_primitive_name, instance_self, hooked_inputs, kwargs, out)

            try:
                out = hook_primitive_outputs(out, captured_grads_output, updated_primitive_name)
            except Exception as exception:
                logger.error(f"This is a primitive op dump error during output hooking: {exception}, "
                             f"primitive_name: {primitive_name}")
                raise DumpException(DumpException.OUTPUT_HOOK_ERROR) from exception

            return out

        return wrapped_primitive_call

    @ThreadSafe.synchronized
    def update_primitive_counters(self, primitive_name):
        if primitive_name not in self.primitive_counters:
            self.primitive_counters[primitive_name] = 0
        else:
            self.primitive_counters[primitive_name] += 1



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
convert format from FRACTAL_NZ to NHWC.
"""

from functools import reduce
import numpy as np


def convert(shape_from: list, shape_to: list, array: any) -> any:
    """
    Convert the data format from FRACTAL_NZ to NHWC
    :param shape_from: the shape before convert
    :param shape_to: the shape after convert
    :param array: the one-dimensional array
    :return: the data array of NHWC shape
    """
    if len(shape_to) == 1:
        axis_h, axis_n, axis_c = 1, 1, shape_to[0]
    elif len(shape_to) == 2:
        axis_h, axis_n, axis_c = 1, shape_to[0], shape_to[1]
    else:
        axis_h, axis_n, axis_c = reduce(lambda x, y: x * y, shape_to[:-2]), shape_to[-2], shape_to[-1]
    axis_ni = shape_from[-2]
    axis_no = shape_from[-3]
    axis_c0 = shape_from[-1]
    axis_c1 = shape_from[-4]
    n_pad = None if axis_no * axis_ni == axis_n else axis_n - axis_no * axis_ni
    c_pad = None if axis_c1 * axis_c0 == axis_c else axis_c - axis_c1 * axis_c0
    tmp_input_tensor = array.reshape(axis_h, axis_c1, axis_no, axis_ni, axis_c0)
    tmp_input_tensor = np.transpose(tmp_input_tensor, axes=(0, 2, 3, 1, 4))
    tmp_input_tensor = tmp_input_tensor.reshape((axis_h, axis_no * axis_ni, axis_c1 * axis_c0))
    data_y = tmp_input_tensor[:, :n_pad, :c_pad]
    if len(shape_to) <= 2:
        data_y = data_y.reshape([data_y.shape[1], data_y.shape[2]])
    return data_y



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
AccumulatedRelativeError algorithm. This file mainly involves the compare function.
"""

import numpy as np

from msprobe.msaccucmp.algorithm_manager.algorithm_parameter import AlgorithmParameter
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager
from msprobe.msaccucmp.cmp_utils import utils


def compare(my_output_dump_data: any, ground_truth_dump_data: any, args: AlgorithmParameter) -> (str, str):
    """
    compare my output dump data and the ground truth dump data
    by accumulated relative error
    formula is (|x[i]-y[i]|) / |y[i]| + (x[i+1]-y[i+1]) / y[i+1]
    :param my_output_dump_data: my output dump data to compare
    :param ground_truth_dump_data: the ground truth dump data to compare
    :param args: the algorithm parameter
    :return: the result of accumulated relative error value and error message (the default is "")
    """
    _ = args  # Bypassing parameter is not used
    result = np.where(
        np.abs(ground_truth_dump_data) > ConstManager.FLOAT_EPSILON,
        np.abs(my_output_dump_data / ground_truth_dump_data - 1),  # abs(aa - bb) / abs(bb) -> abs(aa / bb - 1)
        0,
    ).sum()
    return utils.format_value(result), ""



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from ..utils.constant import SINGLE, DataType
from .graph_repo_base import GraphRepo

JSON_TYPE = DataType.JSON.value


class GraphRepoVis(GraphRepo):
    
    def __init__(self, graph):
        self.graph = graph
        self.repo_type = JSON_TYPE
    
    # 查询根节点信息
    def query_root_nodes(self, graph_type, rank, step): 
        root_node = {}
        if graph_type == SINGLE:
            root_node_name = self.graph.get('root')
            root_node = self.graph.get('node', {}).get(root_node_name, {})
        else:
            root_node_name = self.graph.get(graph_type, {}).get('root')
            root_node = self.graph.get(graph_type, {}).get('node', {}).get(root_node_name, {})
        root_node['node_name'] = root_node_name
        return root_node

    # 查询所有以当前为父节点的子节点
    def query_sub_nodes(self, node_name, graph_type, rank, step):
        sub_nodes = {}
        graph_nodes = {}
        if graph_type == SINGLE:
            graph_nodes = self.graph.get('node', {})
        else:
            graph_nodes = self.graph.get(graph_type, {}).get('node', {})
        target_node = graph_nodes.get(node_name, {})
    
        target_node_children = target_node.get("subnodes", [])
        for subnode_name in target_node_children:
            node_info = graph_nodes.get(subnode_name, {})
            sub_nodes[subnode_name] = node_info
        return sub_nodes
    
    # 查询当前节点的父节点信息
    def query_up_nodes(self, node_name, graph_type, rank, step):
        graph_nodes = {}
        if graph_type == SINGLE:
            graph_nodes = self.graph.get('node', {})
        else:
            graph_nodes = self.graph.get(graph_type, {}).get('node', {})
        
        # 查询当前节点及其的所有父节点，一直到没有父节点位置{}
        up_nodes = {}
        up_nodes[node_name] = graph_nodes.get(node_name, {})
        parent_node_name = graph_nodes.get(node_name, {}).get("upnode")
        while graph_nodes.get(parent_node_name, None) is not None:
            parent_node = graph_nodes.get(parent_node_name, {})
            up_nodes[parent_node_name] = parent_node
            parent_node_name = parent_node.get("upnode")
        return up_nodes
    
    # 查询当前节点信息
    def query_node_info(self, node_name, graph_type):
        graph_nodes = {}
        if graph_type == SINGLE:
            graph_nodes = self.graph.get('node', {})
        else:
            graph_nodes = self.graph.get(graph_type, {}).get('node', {})
        return graph_nodes.get(node_name, {})
        



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import re
import copy
from dataclasses import dataclass

from msprobe.core.common.const import Const
from msprobe.core.common.file_utils import load_json, load_construct_json
from msprobe.core.common.utils import load_stack_json
from msprobe.core.common.log import logger
from msprobe.visualization.builder.msprobe_adapter import get_input_output
from msprobe.visualization.graph.graph import Graph
from msprobe.visualization.graph.node_op import NodeOp
from msprobe.visualization.utils import GraphConst, update_shared_dict, update_pbar_info
from msprobe.visualization.db_utils import node_to_db, config_to_db


class GraphBuilder:
    backward_pattern = re.compile(r"(\.backward\.)(\d+)$")
    forward_pattern = re.compile(r"(\.forward\.)(\d+)$")
    # 匹配以大写字母开头，后接任意字母，并以Template(结尾，或包含api_template(的字符串
    template_pattern = re.compile(r'\b([A-Z][a-zA-Z]*Template|api_template|api_instance)\(')
    micro_step_dict = {}

    @staticmethod
    def build(construct_path, data_path, stack_path, model_name='DefaultModel', pbar_info=None):
        """
        GraphBuilder的对外提供的构图方法
        Args:
            construct_path: construct.json路径
            data_path: dump.json路径
            stack_path: stack.json路径
            model_name: 模型名字，依赖外部输入
            pbar_info: 进度条对象
        Returns: Graph，代表图的数据结构
        """
        construct_dict, micro_step_dict = load_construct_json(construct_path)
        if not construct_dict:
            logger.error("The content of 'construct.json' is empty, failed to build graph. "
                         "When dumping data, it is necessary to select level L0 or mix in order to "
                         "collect model structure data, that is, the content of 'construct.json' is not empty.")
            raise RuntimeError
        GraphBuilder.micro_step_dict = micro_step_dict
        dump_dict = load_json(data_path)
        stack_dict = load_stack_json(stack_path)
        data_dict = dump_dict.get(GraphConst.DATA_KEY, {})
        graph = Graph(model_name, data_path=dump_dict.get('dump_data_dir', ''), dump_data=data_dict,
                      micro_step_num=micro_step_dict.get(Const.MEGATRON_MICRO_STEP_NUMBER))
        GraphBuilder._init_nodes(graph, construct_dict, data_dict, stack_dict, pbar_info=pbar_info)
        GraphBuilder._handle_recompute(graph)
        GraphBuilder._collect_apis_between_modules(graph)
        GraphBuilder._add_parameters_grad(graph, data_dict)
        return graph

    @staticmethod
    def to_db(filename, config, pbar_info=None):
        config.graph_n.step = config.step
        config.graph_n.rank = config.rank
        config.graph_n.compare_mode = config.compare_mode
        node_to_db(config.graph_n, filename, pbar_info)
        if config.graph_b:
            config.graph_b.data_source = GraphConst.JSON_BENCH_KEY
            config.graph_b.step = config.step
            config.graph_b.rank = config.rank
            config.graph_b.compare_mode = config.compare_mode
            node_to_db(config.graph_b, filename, pbar_info)
        config_to_db(config, filename)

    @staticmethod
    def _handle_backward_upnode_missing(construct_dict, subnode_id, upnode_id):
        """
        如果backward节点的父级节点是null，则尝试从同名的forward节点寻找父级节点
        """
        # 匹配以.backward.后跟一个或多个数字结尾的模式
        if GraphBuilder.backward_pattern.search(subnode_id) and not upnode_id:
            forward_upnode_id = construct_dict.get(GraphBuilder.backward_pattern.sub(r".forward.\2", subnode_id))
            if forward_upnode_id:
                new_upnode_id = GraphBuilder.forward_pattern.sub(r".backward.\2", forward_upnode_id)
                if new_upnode_id in construct_dict:
                    return new_upnode_id
        # 匹配以.backward结尾的节点
        if subnode_id.endswith(Const.SEP + Const.BACKWARD) and not upnode_id:
            forward_upnode_id = construct_dict.get(subnode_id.replace(Const.BACKWARD, Const.FORWARD))
            if forward_upnode_id:
                new_upnode_id = forward_upnode_id.replace(Const.FORWARD, Const.BACKWARD)
                if new_upnode_id in construct_dict:
                    return new_upnode_id
        return upnode_id

    @staticmethod
    def _handle_backward_inplace(construct_dict, sub_node_id, up_node_id):
        """
        如果当前backward节点的父层级信息不等于其父级节点的层级信息，则尝试从同名的forward节点寻找父级节点
        主要针对的场景：inplace层会无法触发backward hook导致反向层级错误

        example:
            正确的层级关系：
                父层：Module.layer4.1.BasicBlock.backward.0的层级信息为Module.layer4.1
                子层：Module.layer4.1.conv2.Conv2d.backward.0的父层级信息为Module.layer4.1

            错误的层级关系：
                父层：Module.layer4.1.relu.ReLU.backward.1的层级信息为Module.layer4.1.relu
                子层：Module.layer4.1.conv2.Conv2d.backward.0的父层级信息为Module.layer4.1
        """
        if GraphBuilder.backward_pattern.search(sub_node_id) and up_node_id:
            sub_split = sub_node_id.split(Const.SEP)
            if len(sub_split) < 5:
                return up_node_id
            up_split = up_node_id.split(Const.SEP)
            if len(up_split) < 4:
                return up_node_id
            sub_node_prefix = Const.SEP.join(sub_split[:-4])
            up_node_prefix = Const.SEP.join(up_split[:-3])
            if sub_node_prefix != up_node_prefix:
                forward_sub_node_id = GraphBuilder.backward_pattern.sub(r".forward.\2", sub_node_id)
                if forward_sub_node_id in construct_dict:
                    forward_up_node_id = construct_dict.get(forward_sub_node_id)
                    # forward_up_node_id ---> null
                    if not forward_up_node_id:
                        return forward_up_node_id
                    new_up_node_id = GraphBuilder.forward_pattern.sub(r".backward.\2", forward_up_node_id)
                    if new_up_node_id in construct_dict:
                        return new_up_node_id
        return up_node_id

    @staticmethod
    def _init_nodes(graph, construct_dict, data_dict, stack_dict, pbar_info=None):
        if pbar_info:
            update_shared_dict(pbar_info.current_stage_dict, pbar_info.task_id, 1)
        for i, (subnode_id, upnode_id) in enumerate(construct_dict.items()):
            upnode_id = GraphBuilder._handle_backward_inplace(construct_dict, subnode_id, upnode_id) if upnode_id \
                else GraphBuilder._handle_backward_upnode_missing(construct_dict, subnode_id, upnode_id)
            if upnode_id:
                upnode_op = NodeOp.get_node_op(upnode_id)
                upnode = GraphBuilder._create_or_get_node(graph, [data_dict, stack_dict], upnode_op, upnode_id)
            else:
                upnode = graph.root
            node_op = NodeOp.get_node_op(subnode_id)
            GraphBuilder._create_or_get_node(graph, [data_dict, stack_dict], node_op, subnode_id, upnode)

            if pbar_info:
                update_pbar_info(pbar_info, i + 1, len(construct_dict))

    @staticmethod
    def _create_or_get_node(graph, data_stack_list, op, name, upnode=None):
        if name in graph.node_map:
            node = graph.get_node(name)
        else:
            graph.add_node(op, name, upnode)
            node = graph.get_node(name)
            node_data = data_stack_list[0].get(name, {})
            node_stack_info = data_stack_list[1].get(name, [])
            # 添加输入输出数据
            input_data, output_data = get_input_output(node_data, node.id)
            # 更新数据
            node.set_input_output(input_data, output_data)
            if GraphConst.BATCH_P2P in name:
                GraphBuilder._extract_batch_p2p_info(node, node_data)
            # 反向节点使用对应前向节点的堆栈信息
            # 模块命名举例：Module.module.module.GPTModel.backward.0; API命名举例：Tensor.permute.1.backward
            if (not node_stack_info and
                    (GraphBuilder.backward_pattern.search(name) or name.endswith(f'{Const.SEP}{Const.BACKWARD}'))):
                forward_node = graph.get_node(
                    # 同名模块全局唯一，无论调用几次堆栈信息都一致，直接使用编号0的同名模块堆栈信息，避免遗漏
                    GraphBuilder.backward_pattern.sub(f'{Const.SEP}{Const.FORWARD}{Const.SEP}0', name)) \
                    if GraphBuilder.backward_pattern.search(name) \
                    else graph.get_node(name.replace(Const.BACKWARD, Const.FORWARD))
                node_stack_info = forward_node.stack_info if forward_node \
                    else ['This backward node cannot find the forward node and cannot retrieve stack information.']
            node.stack_info = node_stack_info
            if GraphBuilder.micro_step_dict:
                node.micro_step_id = GraphBuilder.micro_step_dict.get(node.id, 0)
        # 添加节点
        node.add_upnode(upnode)
        return node

    @staticmethod
    def _is_valid_batch_p2p_output(param_list):
        if not isinstance(param_list, list) or not param_list:
            return False
        if not isinstance(param_list[0], list) or not param_list[0]:
            return False
        return True

    @staticmethod
    def _extract_batch_p2p_info(node, node_data):
        param_list = node_data.get(Const.OUTPUT, [])
        # 数据格式："output": [[{param1}, {param2}, ...]]
        if GraphBuilder._is_valid_batch_p2p_output(param_list):
            for param in param_list[0]:
                if not isinstance(param, dict):
                    continue
                info = {GraphConst.OP: param.get(GraphConst.OP), GraphConst.PEER: param.get(GraphConst.PEER),
                        GraphConst.GROUP_ID: param.get(GraphConst.GROUP_ID)}
                node.batch_p2p_info.append(info)

    @staticmethod
    def _collect_apis_between_modules(graph):
        """
        图首次展开，这些首层节点包含许多module和api，api数量很多导致图被拉得很长严重影响查阅，因此将module之间的apis收集起来成为节点
        Args:
            graph: 模型结构

        Returns: None
        """
        i = 0
        output = []
        node_list = graph.root.subnodes
        while i < len(node_list):
            current_node = node_list[i]

            # 当前节点为api，检查后续是否还有api
            if current_node.op == NodeOp.function_api:
                temp_nodes = [current_node]
                i += 1
                while i < len(node_list) and node_list[i].op == NodeOp.function_api:
                    temp_nodes.append(node_list[i])
                    i += 1

                # 检查api节点是否大于等于2个
                if len(temp_nodes) >= 2:
                    # 创建新节点，将这些api节点放入新节点的subnodes属性
                    node_id = graph.add_node(NodeOp.api_collection, GraphConst.APIS_BETWEEN_MODULES,
                                             id_accumulation=True)
                    api_collection_node = graph.get_node(node_id)
                    api_collection_node.subnodes = temp_nodes
                    # 重新确立父子关系
                    for node in temp_nodes:
                        node.upnode = api_collection_node
                    api_collection_node.upnode = graph.root
                    output.append(api_collection_node)
                    if temp_nodes[0].micro_step_id is not None:
                        api_collection_node.micro_step_id = temp_nodes[0].micro_step_id
                else:
                    # 如果连续的api节点不足2个，将它们原样添加到输出列表
                    output.extend(temp_nodes)
            else:
                # 如果当前节点为module，直接添加到输出列表
                output.append(current_node)
                i += 1

        graph.root.subnodes = output

    @staticmethod
    def _add_parameters_grad(graph, data_dict):
        """
        将parameters_grad信息添加到graph中，
        对应模块的parameters_grad节点添加到对应模块的最后一次backward节点（backward计数最大）内作为子节点

        例如，graph有节点Module.a.backward.0, Module.a.backward.1, Module.a.backward.2
        则Module.a.parameters_grad添加在Module.a.backward.2内作为子节点
        """
        prefix_suffix_map = {}
        base_suffix = Const.SEP + Const.PARAMS_GRAD
        pattern = re.compile(rf"{re.escape(base_suffix)}(?:\.\d+)?$")
        for node_id in data_dict.keys():
            if node_id not in graph.node_map:
                m = pattern.search(node_id)
                if not m:
                    continue
                prefix = node_id[:m.start()]
                suffix = node_id[m.start():]
                prefix_suffix_map.setdefault(prefix, []).append(suffix)

        max_info = {prefix: 0 for prefix in prefix_suffix_map}

        for key in graph.node_map.keys():
            parts = key.split(Const.SEP)
            if len(parts) > 2 and parts[-2] == Const.BACKWARD:
                num = int(parts[-1])
                prefix = Const.SEP.join(parts[:-2])
                if prefix in max_info and num > max_info[prefix]:
                    max_info[prefix] = num

        for prefix, num in max_info.items():
            node_id = prefix + Const.SEP + Const.BACKWARD + Const.SEP + str(num)
            node = graph.get_node(node_id)
            if node:
                suffix_list = prefix_suffix_map[prefix]
                for suffix in suffix_list:
                    parameters_grad_node_id = graph.add_node(NodeOp.module, prefix + suffix, up_node=node)
                    # 添加输入输出数据
                    node_data = data_dict.get(parameters_grad_node_id, {})
                    input_data, output_data = get_input_output(node_data, parameters_grad_node_id)
                    # 更新数据
                    graph.get_node(parameters_grad_node_id).set_input_output(input_data, output_data)

    @staticmethod
    def _handle_recompute(graph):
        """
        1. 通过_get_recompute_map获得重计算节点映射recompute_map: dict(node_id: node_id_prefix)
        2. 通过_get_no_recompute_map获得非重计算节点映射no_recompute_map: dict(node_id_prefix: list(node_id))
        3. 遍历recompute_map，通过node_id_prefix与no_recompute_map建立连接，通过非重计算节点找到自身的父节点
        """
        recompute_map, recompute_id_map = GraphBuilder._get_recompute_map(graph.root.subnodes)
        if not recompute_map:
            return
        id_prefixes = set(recompute_map.values())
        no_recompute_map = GraphBuilder._get_no_recompute_map(graph, id_prefixes)
        if not no_recompute_map:
            return
        # 拷贝非重计算节点字典用于反向模式
        no_recompute_ids_b = {node_id: list(node_list) for node_id, node_list in no_recompute_map.items()}

        del_indexes = []
        for node_id, id_prefix in recompute_map.items():
            if id_prefix not in no_recompute_map:
                continue
            node_list = no_recompute_map.get(id_prefix) if GraphBuilder.forward_pattern.search(node_id) else \
                no_recompute_ids_b.get(id_prefix)
            if not node_list:
                continue
            no_recompute_node = node_list.pop()
            recompute_node = graph.node_map.get(node_id)
            if not recompute_node:
                continue
            # 通过非重计算forward节点的父节点，找到对应的backward父节点
            new_up_node = graph.node_map.get(
                GraphBuilder.forward_pattern.sub(r".backward.\2", no_recompute_node.upnode.id))
            if not new_up_node:
                continue

            # 更新节点连接关系
            recompute_node.upnode = new_up_node
            new_up_node.subnodes.append(recompute_node)

            del_indexes.append(recompute_id_map.get(node_id))

        # 从后往前删除graph首层中已更新父节点的重计算节点
        del_indexes.sort(reverse=True)
        for index in del_indexes:
            if 0 <= index <= len(graph.root.subnodes):
                del graph.root.subnodes[index]

    @staticmethod
    def _get_recompute_map(node_list: list):
        """
        找到graph首层的重计算层

        return: dict(node_id: node_id_prefix), dict(node_id: index)

        example:
        {Module.0.module.decoder.layers.0.TransformerLayer.forward.4: Module.0.module.decoder.layers.0.TransformerLayer}
        """
        recompute_map = {}
        recompute_id_map = {}
        node_id_set = set([node.id for node in node_list])
        node_id_cache = set()
        for i, node in enumerate(node_list):
            if NodeOp.get_node_op(node.id) != NodeOp.module:
                continue
            id_segments = node.id.split(Const.SEP)
            prefix = Const.SEP.join(id_segments[:-2])
            if node.id in node_id_cache:
                recompute_map[node.id] = prefix
                recompute_id_map[node.id] = i
                continue
            is_recompute = GraphBuilder._is_recompute_node_id(id_segments)
            if not is_recompute:
                continue
            # 重计算层必然是一组对应的前反向节点
            id_segments[-2] = Const.BACKWARD if id_segments[-2] == Const.FORWARD else Const.FORWARD
            relative_node_id = Const.SEP.join(id_segments)
            if relative_node_id in node_id_set:
                recompute_map[node.id] = prefix
                recompute_id_map[node.id] = i
                # 对应节点id放入缓存避免后续重复判断
                node_id_cache.add(relative_node_id)
        return recompute_map, recompute_id_map

    @staticmethod
    def _is_recompute_node_id(id_segments):
        """
        非重计算首层节点命名必然是：Module/Cell.{number(可选)}.module_name.{number(可选)}.class_name.forward/backward.number
        如果不符合，则判断为重计算节点
        """
        if len(id_segments) > 7:
            return True
        if len(id_segments) == 7 and not (id_segments[1].isdigit() and id_segments[3].isdigit()):
            return True
        if len(id_segments) == 6 and not id_segments[1].isdigit():
            return True
        return False

    @staticmethod
    def _get_no_recompute_map(graph, recompute_id_prefixes):
        """
        寻找与重计算层id前缀相同的非重计算forward层，按顺序排列，重计算层按照顺序使用非重计算forward层的父节点对应的backward节点

        return: dict(node_id_prefix: list(node_id))
        """
        no_recompute_map = {}
        for node_id, node in graph.node_map.items():
            if NodeOp.get_node_op(node_id) == NodeOp.module and GraphBuilder.forward_pattern.search(node_id):
                if not node.upnode or node.upnode.id == graph.root.id:
                    continue
                id_prefix = GraphBuilder.forward_pattern.sub('', node_id)
                if id_prefix not in recompute_id_prefixes:
                    continue
                no_recompute_map.setdefault(id_prefix, []).append(node)
        for node_list in no_recompute_map.values():
            # 方便按顺序pop弹出
            node_list.reverse()
        return no_recompute_map


class GraphExportConfig:
    def __init__(self, graph_n, graph_b=None, tool_tip=None, node_colors=None, micro_steps=None, task='',
                 overflow_check=False, compare_mode=None, step=0, rank=0, step_list=None, rank_list=None):
        self.graph_n = graph_n
        self.graph_b = graph_b
        self.tool_tip = tool_tip
        self.node_colors = node_colors
        self.micro_steps = micro_steps
        self.task = task
        self.overflow_check = overflow_check
        self.compare_mode = compare_mode
        self.step = step
        self.rank = rank
        self.step_list = step_list
        self.rank_list = rank_list


@dataclass
class GraphInfo:
    graph: Graph
    construct_path: str
    data_path: str
    stack_path: str


@dataclass
class BuildGraphTaskInfo:
    graph_info_n: GraphInfo
    graph_info_b: GraphInfo
    npu_rank: str
    bench_rank: str
    time_str: str



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import torch


def npu_rotary_mul(x, r1, r2):
    x1, x2 = torch.chunk(x, 2, -1)
    x_new = torch.cat((-x2, x1), dim=-1)
    output = r1 * x + r2 * x_new
    return output


def npu_rotary_mul_backward(dy_tensor, x, r1, r2):
    x.requires_grad = True
    r1.requires_grad = True
    r2.requires_grad = True
    # golden
    x1, x2 = torch.chunk(x, 2, -1)
    x_new = torch.cat((-x2, x1), dim=-1)
    golden_tensor = r1 * x + r2 * x_new
    golden_tensor.backward(dy_tensor)
    r1_shape = r1.shape
    r1_grad = torch.zeros(r1_shape).type(torch.float32)
    r2_grad = torch.zeros(r1_shape).type(torch.float32)
    x1, x2 = torch.chunk(x.float(), 2, -1)
    x_new2 = torch.cat((-x2, x1), dim=-1)
    x_shape = x.shape
    h = x.float()
    grad = dy_tensor.float()
    if len(r1_shape) < 4 or len(x_shape) < 4:
        raise RuntimeError(f"Shape of r1 and x should at least be 4-dimension, "
                           f"but got r1 shape:{r1_shape}, x shape:{x_shape}")
    condition_1 = (r1_shape[0] == 1
                   and r1_shape[1] == x_shape[1]
                   and r1_shape[2] == 1
                   and r1_shape[3] == x_shape[3])
    condition_2 = (r1_shape[0] == 1
                   and r1_shape[1] == 1
                   and r1_shape[2] == x_shape[2]
                   and r1_shape[3] == x_shape[3])
    condition_3 = (r1_shape[0] == x_shape[0]
                   and r1_shape[1] == 1
                   and r1_shape[2] == 1
                   and r1_shape[3] == x_shape[3])

    if condition_1:
        for i in range(x_shape[0]):
            for j in range(x_shape[2]):
                r2_grad[0, :, 0, :] += (x_new2[i, :, j, :] * grad[i, :, j, :])
                r1_grad[0, :, 0, :] += (h[i, :, j, :] * grad[i, :, j, :])
    elif condition_2:
        for i in range(x_shape[0]):
            for j in range(x_shape[1]):
                r2_grad[0, 0, :, :] += (x_new2[i, j, :, :] * grad[i, j, :, :])
                r1_grad[0, 0, :, :] += (h[i, j, :, :] * grad[i, j, :, :])
    elif condition_3:
        for i in range(x_shape[1]):
            for j in range(x_shape[2]):
                r2_grad[:, 0, 0, :] += (x_new2[:, i, j, :] * grad[:, i, j, :])
                r1_grad[:, 0, 0, :] += (h[:, i, j, :] * grad[:, i, j, :])

    return x.grad.cpu(), r1_grad.cpu(), r2_grad.cpu()
